Package: rkaf
Title: Kolmogorov-Arnold Fourier Networks in R
Version: 0.1.0
Authors@R: 
    person(
      given = "Guillaume",
      family = "Sidoine",
      email = "gjh.sidoine@gmail.com",
      role = c("aut", "cre")
    )
Description: Provides an R implementation of Kolmogorov-Arnold Fourier 
    Networks using the torch backend. The package supports regression, binary 
    classification, multiclass classification, formula and matrix interfaces, 
    mini-batch training, validation splits, early stopping, standardization, 
    best-model restoration, and KAF-specific diagnostics.
License: MIT + file LICENSE
Encoding: UTF-8
RoxygenNote: 7.3.3
Depends: R (>= 4.1.0)
Imports: stats, graphics, torch
Suggests: testthat (>= 3.0.0), knitr, rmarkdown, pkgdown
Config/testthat/edition: 3
URL: https://github.com/gsidoine/rkaf, https://gsidoine.github.io/rkaf/
BugReports: https://github.com/gsidoine/rkaf/issues
VignetteBuilder: knitr
NeedsCompilation: no
Packaged: 2026-04-26 17:46:14 UTC; gjhsi
Author: Guillaume Sidoine [aut, cre]
Maintainer: Guillaume Sidoine <gjh.sidoine@gmail.com>
Repository: CRAN
Date/Publication: 2026-04-28 19:40:02 UTC
Built: R 4.6.0; ; 2026-05-01 21:22:19 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method(plot,kaf_fit)
S3method(predict,kaf_fit)
S3method(print,kaf_fit)
export(extract_fourier_params)
export(extract_kaf_scales)
export(kaf)
export(kaf_fit)
export(kaf_fit_formula)
export(nn_kaf)
export(nn_kaf_layer)
export(nn_random_fourier_features)
export(plot_kaf_scales)

<!-- README.md is generated from README.Rmd. Please edit README.Rmd -->

# rkaf

`rkaf` is an R package for **Kolmogorov-Arnold Fourier Networks** using
the [`torch`](https://torch.mlverse.org/) backend.

The package provides a modern R interface for KAF models, including:

- regression
- binary classification
- multiclass classification
- formula and matrix interfaces
- mini-batch training
- validation splits
- early stopping
- automatic standardization
- best-model restoration
- KAF-specific diagnostics

The goal of `rkaf` is to make KAF-style neural networks accessible to R
users without requiring Python, `reticulate`, or custom training loops.

## Installation

You can install the development version locally with:

``` r
devtools::install()
```

Or from a local package directory:

``` r
devtools::load_all()
```

`rkaf` depends on `torch`. If `torch` is not already configured, run:

``` r
torch::install_torch()
```

## Basic regression example

``` r
library(rkaf)

set.seed(123)
torch::torch_manual_seed(123)

x <- as.matrix(seq(-1, 1, length.out = 128))

y <- sin(8 * pi * x) +
  0.35 * cos(3 * pi * x) +
  0.15 * x^2

fit <- kaf_fit(
  x = x,
  y = y,
  hidden = c(256, 256),
  num_grids = 32,
  use_layernorm = FALSE,
  epochs = 1000,
  lr = 1e-3,
  standardize_x = FALSE,
  standardize_y = TRUE,
  fourier_init_scale = 5e-2,
  restore_best = TRUE,
  verbose = FALSE,
  seed = 123
)

fit
#> <kaf_fit>
#> Task:         regression
#> Architecture: 1 -> 256 -> 256 -> 1
#> Fourier grids: 32 
#> Epochs:       1000
#> Batch size:   128
#> Validation:   no
#> Standardize x: no 
#> Standardize y: yes 
#> Final train loss: 0.0227977
#> Best train loss:  0.0198582 at epoch 945
```

``` r
pred <- predict(fit, x)

plot(
  x,
  y,
  type = "l",
  lwd = 2,
  xlab = "x",
  ylab = "f(x)",
  main = "KAF regression example"
)

lines(x, pred, lwd = 2, lty = 2)

legend(
  "topright",
  legend = c("Observed", "Predicted"),
  lty = c(1, 2),
  lwd = 2,
  bty = "n"
)
```

<img src="man/figures/README-regression-prediction-plot-1.png" alt="" width="100%" />

``` r
plot(fit)
```

<img src="man/figures/README-regression-loss-plot-1.png" alt="" width="100%" />

This example intentionally uses a Fourier-heavy target function to
demonstrate the model’s ability to learn oscillatory structure.

## Formula interface

`rkaf` also supports a formula interface for tabular data.

``` r
fit_mtcars <- kaf_fit_formula(
  mpg ~ wt + hp + cyl,
  data = mtcars,
  hidden = c(32, 32),
  num_grids = 16,
  epochs = 200,
  verbose = FALSE,
  seed = 123
)

fit_mtcars
#> <kaf_fit>
#> Task:         regression
#> Formula:      mpg ~ wt + hp + cyl
#> Architecture: 3 -> 32 -> 32 -> 1
#> Fourier grids: 16 
#> Epochs:       200
#> Batch size:   32
#> Validation:   no
#> Standardize x: yes 
#> Standardize y: yes 
#> Final train loss: 0.0542788
#> Best train loss:  0.0542788 at epoch 200
```

``` r
mtcars_pred <- predict(fit_mtcars, mtcars)

head(data.frame(
  observed = mtcars$mpg,
  predicted = round(mtcars_pred, 2)
))
#>   observed predicted
#> 1     21.0     21.16
#> 2     21.0     21.15
#> 3     22.8     21.86
#> 4     21.4     20.03
#> 5     18.7     16.58
#> 6     18.1     18.30
```

## Binary classification

For factor, character, or logical targets with two classes, `rkaf` can
fit a binary classifier automatically.

``` r
df <- mtcars
df$high_mpg <- factor(
  ifelse(df$mpg > median(df$mpg), "yes", "no"),
  levels = c("no", "yes")
)

fit_binary <- kaf_fit_formula(
  high_mpg ~ wt + hp + cyl,
  data = df,
  hidden = c(32, 32),
  num_grids = 16,
  epochs = 200,
  verbose = FALSE,
  seed = 123
)

fit_binary
#> <kaf_fit>
#> Task:         binary
#> Classes:      no, yes
#> Formula:      high_mpg ~ wt + hp + cyl
#> Architecture: 3 -> 32 -> 32 -> 1
#> Fourier grids: 16 
#> Epochs:       200
#> Batch size:   32
#> Validation:   no
#> Standardize x: yes 
#> Standardize y: no 
#> Final train loss: 0.00163561
#> Best train loss:  0.00163561 at epoch 200
```

Predicted probabilities and class:

``` r
prob_binary <- predict(fit_binary, df, type = "prob")
class_binary <- predict(fit_binary, df, type = "class")

head(data.frame(
  observed = df$high_mpg,
  prob_yes = round(prob_binary, 3),
  predicted = class_binary
))
#>   observed prob_yes predicted
#> 1      yes    0.999       yes
#> 2      yes    0.999       yes
#> 3      yes    1.000       yes
#> 4      yes    0.983       yes
#> 5       no    0.000        no
#> 6       no    0.008        no

table(
  observed = df$high_mpg,
  predicted = class_binary
)
#>         predicted
#> observed no yes
#>      no  17   0
#>      yes  0  15
```

## Multiclass classification

For factor targets with more than two classes, `rkaf` fits a multiclass
model.

``` r
fit_iris <- kaf_fit_formula(
  Species ~ .,
  data = iris,
  hidden = c(32, 32),
  num_grids = 16,
  epochs = 300,
  verbose = FALSE,
  seed = 123
)

fit_iris
#> <kaf_fit>
#> Task:         multiclass
#> Classes:      setosa, versicolor, virginica
#> Formula:      Species ~ .
#> Architecture: 4 -> 32 -> 32 -> 3
#> Fourier grids: 16 
#> Epochs:       300
#> Batch size:   150
#> Validation:   no
#> Standardize x: yes 
#> Standardize y: no 
#> Final train loss: 0.00819381
#> Best train loss:  0.00819381 at epoch 300
```

Class probabilities:

``` r
head(round(predict(fit_iris, iris, type = "prob"), 3))
#>      setosa versicolor virginica
#> [1,]  0.999      0.001     0.000
#> [2,]  0.995      0.003     0.002
#> [3,]  0.997      0.001     0.001
#> [4,]  0.997      0.001     0.001
#> [5,]  0.999      0.000     0.000
#> [6,]  1.000      0.000     0.000
```

Predicted classes:

``` r
class_iris <- predict(fit_iris, iris, type = "class")

table(
  observed = iris$Species,
  predicted = class_iris
)
#>             predicted
#> observed     setosa versicolor virginica
#>   setosa         50          0         0
#>   versicolor      0         50         0
#>   virginica       0          0        50
```

## Validation and early stopping

`kaf_fit()` supports validation splits, mini-batching, and early
stopping.

``` r
fit_val <- kaf_fit(
  x = x,
  y = y,
  hidden = c(64, 64),
  num_grids = 16,
  use_layernorm = FALSE,
  epochs = 300,
  lr = 5e-4,
  batch_size = 64,
  validation_split = 0.2,
  patience = 100,
  restore_best = TRUE,
  verbose = FALSE,
  seed = 123
)

fit_val$train_loss_history
fit_val$validation_loss_history
```

The fitted object stores both training and validation loss histories, so
users can inspect them directly through `fit_val$train_loss_history` and
`fit_val$validation_loss_history`. The code above is shown as an example
and is not run when building this README.

## KAF diagnostics

You can inspect the learned balance between the base/GELU branch and the
Fourier branch.

``` r
scales <- extract_kaf_scales(fit)
head(scales)
#>   layer feature base_scale fourier_scale fourier_to_base_ratio
#> 1     1       1  1.0369691    0.11137812            0.10740737
#> 2     2       1  0.9652719    0.15750510            0.16317174
#> 3     2       2  0.9738784    0.39944243            0.41015638
#> 4     2       3  0.9945174    0.12360517            0.12428658
#> 5     2       4  0.9936960    0.08714252            0.08769535
#> 6     2       5  1.0083405    0.08873361            0.08799965
```

``` r
fourier_params <- extract_fourier_params(fit, layer = 1)
head(fourier_params)
#>   layer input_feature grid     weight      bias
#> 1     1             1    1  0.1379520 4.5462580
#> 2     1             1    2 -0.1811931 4.3979001
#> 3     1             1    3 -0.1297648 5.7381864
#> 4     1             1    4 -0.4164377 2.7132401
#> 5     1             1    5  0.3679259 0.4899397
#> 6     1             1    6  0.3606938 2.3420620
```

## Low-level torch interface

Advanced users can directly create KAF torch modules.

``` r
model <- nn_kaf(
  layers = c(4, 16, 16, 1),
  num_grids = 8
)

x_tensor <- torch::torch_randn(10, 4)
y_tensor <- model(x_tensor)

y_tensor$shape
#> [1] 10  1
```

## Current package status

The package currently includes:

``` r
devtools::test()
devtools::check()
```

with full test coverage for the core API, including regression, binary
classification, multiclass classification, formula handling, prediction,
training utilities, and diagnostics.

## References and attribution

`rkaf` is an independent R implementation of Kolmogorov-Arnold Fourier
Networks using the `torch` backend.

The package is based on the KAF architecture proposed in:

> Zhang, J., Fan, Y., Cai, K., & Wang, K. (2025).  
> Kolmogorov-Arnold Fourier Networks.  
> arXiv:2502.06018.

The implementation was also informed by the public Python reference
implementations:

- `kolmogorovArnoldFourierNetwork/KAF`
- `kolmogorovArnoldFourierNetwork/kaf_act`

No Python dependency is required by `rkaf`.

## License

MIT.

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  message = FALSE,
  warning = FALSE
)

rkaf_torch_available <- tryCatch(
  {
    torch::torch_manual_seed(123)
    invisible(torch::torch_tensor(0))
    TRUE
  },
  error = function(e) FALSE
)

knitr::opts_chunk$set(eval = rkaf_torch_available)

## ----torch-unavailable, echo = FALSE, eval = !rkaf_torch_available------------
# knitr::asis_output(
#   "> Torch/Lantern is not available in this build environment, so the code chunks are shown but not executed."
# )

## ----setup--------------------------------------------------------------------
library(rkaf)

set.seed(123)
torch::torch_manual_seed(123)

## ----regression-data----------------------------------------------------------
x <- as.matrix(seq(-1, 1, length.out = 128))

y <- sin(8 * pi * x) +
  0.35 * cos(3 * pi * x) +
  0.15 * x^2

## ----regression-fit-----------------------------------------------------------
fit <- kaf_fit(
  x = x,
  y = y,
  hidden = c(256, 256),
  num_grids = 32,
  use_layernorm = FALSE,
  epochs = 1000,
  lr = 1e-3,
  standardize_x = FALSE,
  standardize_y = TRUE,
  fourier_init_scale = 5e-2,
  restore_best = TRUE,
  verbose = FALSE,
  seed = 123
)

fit

## ----regression-predict-------------------------------------------------------
pred <- predict(fit, x)

head(data.frame(
  observed = round(as.numeric(y), 3),
  predicted = round(pred, 3)
))

## ----regression-plot, fig.width=7, fig.height=4-------------------------------
plot(
  x,
  y,
  type = "l",
  lwd = 2,
  xlab = "x",
  ylab = "f(x)",
  main = "KAF regression fit"
)

lines(x, pred, lwd = 2, lty = 2)

legend(
  "topright",
  legend = c("Observed", "Predicted"),
  lty = c(1, 2),
  lwd = 2,
  bty = "n"
)

## ----formula-regression-------------------------------------------------------
fit_mtcars <- kaf_fit_formula(
  mpg ~ wt + hp + cyl,
  data = mtcars,
  hidden = c(32, 32),
  num_grids = 16,
  epochs = 200,
  verbose = FALSE,
  seed = 123
)

fit_mtcars

## ----formula-predict----------------------------------------------------------
mtcars_pred <- predict(fit_mtcars, mtcars)

head(data.frame(
  observed = mtcars$mpg,
  predicted = round(mtcars_pred, 2)
))

## ----binary-data--------------------------------------------------------------
df <- mtcars

df$high_mpg <- factor(
  ifelse(df$mpg > median(df$mpg), "yes", "no"),
  levels = c("no", "yes")
)

## ----binary-fit---------------------------------------------------------------
fit_binary <- kaf_fit_formula(
  high_mpg ~ wt + hp + cyl,
  data = df,
  hidden = c(32, 32),
  num_grids = 16,
  epochs = 200,
  verbose = FALSE,
  seed = 123
)

fit_binary

## ----binary-prob--------------------------------------------------------------
prob_binary <- predict(fit_binary, df, type = "prob")
class_binary <- predict(fit_binary, df, type = "class")

head(data.frame(
  observed = df$high_mpg,
  prob_yes = round(prob_binary, 3),
  predicted = class_binary
))

## ----binary-confusion-matrix--------------------------------------------------
table(
  observed = df$high_mpg,
  predicted = class_binary
)

## ----binary-link--------------------------------------------------------------
head(predict(fit_binary, df, type = "link"))

## ----multiclass-fit-----------------------------------------------------------
fit_iris <- kaf_fit_formula(
  Species ~ .,
  data = iris,
  hidden = c(32, 32),
  num_grids = 16,
  epochs = 300,
  verbose = FALSE,
  seed = 123
)

fit_iris

## ----multiclass-confusion-matrix----------------------------------------------
class_iris <- predict(fit_iris, iris, type = "class")

table(
  observed = iris$Species,
  predicted = class_iris
)

## ----multiclass-prob----------------------------------------------------------
head(round(predict(fit_iris, iris, type = "prob"), 3))

## ----validation-fit, eval=FALSE-----------------------------------------------
# fit_val <- kaf_fit(
#   x = x,
#   y = y,
#   hidden = c(64, 64),
#   num_grids = 16,
#   use_layernorm = FALSE,
#   epochs = 300,
#   lr = 5e-4,
#   batch_size = 64,
#   validation_split = 0.2,
#   patience = 100,
#   restore_best = TRUE,
#   verbose = FALSE,
#   seed = 123
# )
# 
# plot(fit_val)

## ----diagnostics-scales-------------------------------------------------------
scales <- extract_kaf_scales(fit)

head(scales)

## ----diagnostics-fourier------------------------------------------------------
fourier_params <- extract_fourier_params(fit, layer = 1)

head(fourier_params)

## ----low-level----------------------------------------------------------------
model <- nn_kaf(
  layers = c(4, 16, 16, 1),
  num_grids = 8
)

x_tensor <- torch::torch_randn(10, 4)
y_tensor <- model(x_tensor)

y_tensor$shape

## ----workflow, eval=FALSE-----------------------------------------------------
# fit <- kaf_fit_formula(
#   y ~ .,
#   data = df,
#   hidden = c(64, 64),
#   num_grids = 16,
#   validation_split = 0.2,
#   patience = 30
# )
# 
# predict(fit, newdata)
# plot(fit)
# extract_kaf_scales(fit)

## ----classification-workflow, eval=FALSE--------------------------------------
# predict(fit, newdata, type = "prob")
# predict(fit, newdata, type = "class")

#' Fit a KAF model using an R formula
#'
#' Fits a Kolmogorov-Arnold Fourier Network using a formula and data frame.
#' This is a convenience wrapper around `kaf_fit()` for regression tasks.
#'
#' @param formula A model formula, such as `y ~ x1 + x2`.
#' @param data A data frame.
#' @param include_intercept Logical. Whether to keep the model-matrix intercept
#'   column. Defaults to `FALSE`.
#' @param na.action Missing-value handling function passed to `model.frame()`.
#' @param ... Additional arguments passed to `kaf_fit()`.
#'
#' @return An object of class `"kaf_fit"`.
#'
#' @export
kaf_fit_formula <- function(formula,
                            data,
                            include_intercept = FALSE,
                            na.action = stats::na.omit,
                            ...) {
  if (!inherits(formula, "formula")) {
    stop("`formula` must be a formula.", call. = FALSE)
  }

  if (!is.data.frame(data)) {
    stop("`data` must be a data frame.", call. = FALSE)
  }

  mf <- stats::model.frame(
    formula = formula,
    data = data,
    na.action = na.action
  )

  y <- stats::model.response(mf)

  terms_full <- stats::terms(formula, data = data)
  x_terms <- stats::delete.response(terms_full)

  x <- stats::model.matrix(x_terms, data = mf)

  if (!isTRUE(include_intercept) && "(Intercept)" %in% colnames(x)) {
    x <- x[, colnames(x) != "(Intercept)", drop = FALSE]
  }

  fit <- kaf_fit(
    x = x,
    y = y,
    ...
  )

  fit$formula <- formula
  fit$terms <- x_terms
  fit$xlevels <- stats::.getXlevels(x_terms, mf)
  fit$contrasts <- attr(x, "contrasts")
  fit$x_colnames <- colnames(x)
  fit$include_intercept <- include_intercept
  fit$response_name <- deparse(formula[[2]])

  fit
}


model_matrix_from_kaf_formula <- function(object, newdata) {
  if (!is.data.frame(newdata)) {
    stop("`newdata` must be a data frame for formula-based KAF fits.",
         call. = FALSE)
  }

  mf <- stats::model.frame(
    formula = object$terms,
    data = newdata,
    na.action = stats::na.pass,
    xlev = object$xlevels
  )

  x <- stats::model.matrix(
    object$terms,
    data = mf,
    contrasts.arg = object$contrasts
  )

  if (!isTRUE(object$include_intercept) && "(Intercept)" %in% colnames(x)) {
    x <- x[, colnames(x) != "(Intercept)", drop = FALSE]
  }

  missing_cols <- setdiff(object$x_colnames, colnames(x))

  if (length(missing_cols) > 0) {
    stop(
      "`newdata` is missing model matrix columns: ",
      paste(missing_cols, collapse = ", "),
      call. = FALSE
    )
  }

  x[, object$x_colnames, drop = FALSE]
}

test_that("kaf_fit() supports binary classification with factor targets", {
  skip_if_no_torch()
  torch::torch_manual_seed(123)

  x <- matrix(seq(-1, 1, length.out = 80), ncol = 1)
  y <- factor(ifelse(x[, 1] > 0, "yes", "no"))

  fit <- kaf_fit(
    x = x,
    y = y,
    task = "binary",
    hidden = c(8),
    num_grids = 4,
    use_layernorm = FALSE,
    epochs = 5,
    verbose = FALSE,
    seed = 123
  )

  prob <- predict(fit, x, type = "prob")
  cls <- predict(fit, x, type = "class")
  logits <- predict(fit, x, type = "link")

  expect_s3_class(fit, "kaf_fit")
  expect_equal(fit$task, "binary")
  expect_equal(fit$output_dim, 1)
  expect_equal(fit$class_levels, c("no", "yes"))

  expect_type(prob, "double")
  expect_true(all(prob >= 0 & prob <= 1))
  expect_equal(length(prob), nrow(x))

  expect_s3_class(cls, "factor")
  expect_equal(levels(cls), c("no", "yes"))

  expect_type(logits, "double")
  expect_equal(length(logits), nrow(x))
})


test_that("kaf_fit() supports multiclass classification with factor targets", {
  skip_if_no_torch()
  torch::torch_manual_seed(123)

  x <- cbind(
    x1 = rnorm(90),
    x2 = rnorm(90)
  )

  y <- factor(rep(c("a", "b", "c"), each = 30))

  fit <- kaf_fit(
    x = x,
    y = y,
    task = "multiclass",
    hidden = c(8),
    num_grids = 4,
    epochs = 5,
    verbose = FALSE,
    seed = 123
  )

  prob <- predict(fit, x, type = "prob")
  cls <- predict(fit, x, type = "class")
  logits <- predict(fit, x, type = "link")

  expect_s3_class(fit, "kaf_fit")
  expect_equal(fit$task, "multiclass")
  expect_equal(fit$output_dim, 3)
  expect_equal(fit$class_levels, c("a", "b", "c"))

  expect_true(is.matrix(prob))
  expect_equal(dim(prob), c(90L, 3L))
  expect_equal(colnames(prob), c("a", "b", "c"))

  expect_s3_class(cls, "factor")
  expect_equal(levels(cls), c("a", "b", "c"))

  expect_true(is.matrix(logits))
  expect_equal(dim(logits), c(90L, 3L))
})


test_that("kaf_fit() auto-detects factor targets as classification", {
  skip_if_no_torch()
  torch::torch_manual_seed(123)

  x <- matrix(seq(-1, 1, length.out = 60), ncol = 1)
  y <- factor(ifelse(x[, 1] > 0, "positive", "negative"))

  fit <- kaf_fit(
    x = x,
    y = y,
    hidden = c(8),
    num_grids = 4,
    epochs = 5,
    verbose = FALSE,
    seed = 123
  )

  expect_equal(fit$task, "binary")
})


test_that("kaf_fit() rejects standardize_y for classification", {
  skip_if_no_torch()
  x <- matrix(seq(-1, 1, length.out = 20), ncol = 1)
  y <- factor(ifelse(x[, 1] > 0, "yes", "no"))

  expect_error(
    kaf_fit(
      x = x,
      y = y,
      task = "binary",
      standardize_y = TRUE,
      hidden = c(4),
      epochs = 2,
      verbose = FALSE
    ),
    "`standardize_y = TRUE` is only supported for regression"
  )
})


test_that("formula interface supports binary classification", {
  skip_if_no_torch()
  torch::torch_manual_seed(123)

  df <- data.frame(
    y = factor(rep(c("no", "yes"), each = 30)),
    x1 = rnorm(60),
    x2 = rnorm(60)
  )

  fit <- kaf_fit_formula(
    y ~ x1 + x2,
    data = df,
    hidden = c(8),
    num_grids = 4,
    epochs = 5,
    verbose = FALSE,
    seed = 123
  )

  prob <- predict(fit, df, type = "prob")
  cls <- predict(fit, df, type = "class")

  expect_equal(fit$task, "binary")
  expect_type(prob, "double")
  expect_equal(length(prob), nrow(df))
  expect_s3_class(cls, "factor")
})


test_that("formula interface supports multiclass classification", {
  skip_if_no_torch()
  torch::torch_manual_seed(123)

  df <- data.frame(
    y = factor(rep(c("a", "b", "c"), each = 20)),
    x1 = rnorm(60),
    x2 = rnorm(60)
  )

  fit <- kaf_fit_formula(
    y ~ x1 + x2,
    data = df,
    hidden = c(8),
    num_grids = 4,
    epochs = 5,
    verbose = FALSE,
    seed = 123
  )

  prob <- predict(fit, df, type = "prob")
  cls <- predict(fit, df, type = "class")

  expect_equal(fit$task, "multiclass")
  expect_true(is.matrix(prob))
  expect_equal(dim(prob), c(60L, 3L))
  expect_s3_class(cls, "factor")
})

test_that("kaf_fit() supports multiclass classification with validation split", {
  skip_if_no_torch()
  torch::torch_manual_seed(123)

  x <- cbind(
    x1 = rnorm(90),
    x2 = rnorm(90)
  )

  y <- factor(rep(c("a", "b", "c"), each = 30))

  fit <- kaf_fit(
    x = x,
    y = y,
    task = "multiclass",
    hidden = c(8),
    num_grids = 4,
    epochs = 5,
    validation_split = 0.2,
    verbose = FALSE,
    seed = 123
  )

  prob <- predict(fit, x, type = "prob")

  expect_s3_class(fit, "kaf_fit")
  expect_equal(fit$task, "multiclass")
  expect_true(any(!is.na(fit$validation_loss_history)))
  expect_true(is.matrix(prob))
  expect_equal(ncol(prob), 3L)
})

test_that("kaf_fit_formula() fits a numeric regression formula", {
  skip_if_no_torch()
  torch::torch_manual_seed(123)

  df <- data.frame(
    y = sin(seq(-1, 1, length.out = 40)),
    x = seq(-1, 1, length.out = 40),
    z = seq(0, 1, length.out = 40)
  )

  fit <- kaf_fit_formula(
    y ~ x + z,
    data = df,
    hidden = c(8),
    num_grids = 4,
    use_layernorm = FALSE,
    epochs = 5,
    verbose = FALSE,
    seed = 123
  )

  expect_s3_class(fit, "kaf_fit")
  expect_equal(fit$input_dim, 2)
  expect_equal(fit$response_name, "y")
  expect_false(is.null(fit$formula))
})


test_that("predict() works with formula-based KAF fits and data frames", {
  skip_if_no_torch()
  torch::torch_manual_seed(123)

  df <- data.frame(
    y = sin(seq(-1, 1, length.out = 40)),
    x = seq(-1, 1, length.out = 40),
    z = seq(0, 1, length.out = 40)
  )

  fit <- kaf_fit_formula(
    y ~ x + z,
    data = df,
    hidden = c(8),
    num_grids = 4,
    use_layernorm = FALSE,
    epochs = 5,
    verbose = FALSE,
    seed = 123
  )

  pred <- predict(fit, df)

  expect_type(pred, "double")
  expect_equal(length(pred), nrow(df))
})


test_that("kaf_fit_formula() supports factor predictors", {
  skip_if_no_torch()
  torch::torch_manual_seed(123)

  df <- data.frame(
    y = rnorm(60),
    x = rnorm(60),
    group = factor(rep(c("a", "b", "c"), each = 20))
  )

  fit <- kaf_fit_formula(
    y ~ x + group,
    data = df,
    hidden = c(8),
    num_grids = 4,
    use_layernorm = FALSE,
    epochs = 5,
    verbose = FALSE,
    seed = 123
  )

  pred <- predict(fit, df)

  expect_type(pred, "double")
  expect_equal(length(pred), nrow(df))
  expect_true(fit$input_dim >= 3)
})


test_that("kaf_fit_formula() auto-detects factor targets as classification", {
  skip_if_no_torch()
  torch::torch_manual_seed(123)

  df <- data.frame(
    y = factor(rep(c("a", "b"), each = 20)),
    x = rnorm(40)
  )

  fit <- kaf_fit_formula(
    y ~ x,
    data = df,
    hidden = c(4),
    num_grids = 4,
    epochs = 2,
    verbose = FALSE,
    seed = 123
  )

  pred <- predict(fit, df, type = "class")

  expect_s3_class(fit, "kaf_fit")
  expect_equal(fit$task, "binary")
  expect_s3_class(pred, "factor")
  expect_equal(levels(pred), c("a", "b"))
})

#' Create a Kolmogorov-Arnold Fourier Network
#'
#' Convenience wrapper around `nn_kaf()` using a more user-facing API.
#'
#' @param input_dim Integer. Number of input features.
#' @param output_dim Integer. Number of output dimensions.
#' @param hidden Integer vector. Hidden layer sizes.
#' @param num_grids Integer. Number of Fourier frequencies per KAF layer.
#' @param dropout Numeric. Dropout probability.
#' @param use_layernorm Logical. Whether to apply layer normalization.
#' @param activation_expectation Numeric. Scaling constant for Fourier initialization.
#' @param fourier_init_scale Numeric. Initial scale of the Fourier branch.
#'
#' @return A torch KAF network.
#'
#' @export
kaf <- function(input_dim,
                output_dim = 1,
                hidden = c(64, 64),
                num_grids = 16,
                dropout = 0,
                use_layernorm = TRUE,
                activation_expectation = 1.64,
                fourier_init_scale = 1e-2) {
  if (!is.numeric(input_dim) || length(input_dim) != 1 || input_dim < 1) {
    stop("`input_dim` must be a positive integer.", call. = FALSE)
  }

  if (!is.numeric(output_dim) || length(output_dim) != 1 || output_dim < 1) {
    stop("`output_dim` must be a positive integer.", call. = FALSE)
  }

  if (!is.numeric(hidden) || any(hidden < 1)) {
    stop("`hidden` must be a numeric vector of positive integers.", call. = FALSE)
  }

  layers <- as.integer(c(input_dim, hidden, output_dim))

  nn_kaf(
    layers = layers,
    num_grids = num_grids,
    dropout = dropout,
    use_layernorm = use_layernorm,
    activation_expectation = activation_expectation,
    fourier_init_scale = fourier_init_scale
  )
}

#' Kolmogorov-Arnold Fourier Network
#'
#' Torch module implementing a stacked Kolmogorov-Arnold Fourier Network.
#'
#' @param layers Integer vector. Network architecture, including input and
#'   output dimensions. For example, `c(10, 64, 64, 1)`.
#' @param num_grids Integer. Number of Fourier frequencies per KAF layer.
#' @param dropout Numeric. Dropout probability.
#' @param use_layernorm Logical. Whether to apply layer normalization before
#'   the Fourier feature block.
#' @param activation_expectation Numeric. Scaling constant for Fourier
#'   initialization.
#' @param fourier_init_scale Numeric. Initial scale of the Fourier component.
#'
#' @return A torch nn_module.
#'
#' @export
nn_kaf <- torch::nn_module(
  "nn_kaf",

  initialize = function(layers,
                        num_grids = 8,
                        dropout = 0,
                        use_layernorm = TRUE,
                        activation_expectation = 1.64,
                        fourier_init_scale = 1e-2) {
    if (!is.numeric(layers) || length(layers) < 2) {
      stop("`layers` must be a numeric vector with at least two entries.",
           call. = FALSE)
    }

    if (any(layers < 1)) {
      stop("All entries in `layers` must be positive integers.",
           call. = FALSE)
    }

    self$layers_spec <- as.integer(layers)
    self$num_layers <- length(self$layers_spec) - 1

    self$kaf_layers <- torch::nn_module_list()

    for (i in seq_len(self$num_layers)) {
      self$kaf_layers$append(
        nn_kaf_layer(
          input_dim = self$layers_spec[[i]],
          output_dim = self$layers_spec[[i + 1]],
          num_grids = num_grids,
          dropout = dropout,
          use_layernorm = use_layernorm,
          activation_expectation = activation_expectation,
          fourier_init_scale = fourier_init_scale
        )
      )
    }
  },

  forward = function(x) {
    for (i in seq_len(self$num_layers)) {
      x <- self$kaf_layers[[i]](x)
    }

    x
  }
)

tensor_to_numeric <- function(x) {
  as.numeric(x$detach())
}


#' Extract KAF branch scales
#'
#' Extracts the learned GELU/base and Fourier branch scales from each KAF layer.
#'
#' @param object A fitted object returned by `kaf_fit()`, or a raw `nn_kaf`
#'   torch module.
#'
#' @return A data frame with one row per layer-feature pair.
#'
#' @export
extract_kaf_scales <- function(object) {
  model <- if (inherits(object, "kaf_fit")) object$model else object

  if (is.null(model$kaf_layers) || is.null(model$num_layers)) {
    stop("`object` must be a fitted KAF object or an `nn_kaf` model.",
         call. = FALSE)
  }

  out <- vector("list", model$num_layers)

  for (i in seq_len(model$num_layers)) {
    layer <- model$kaf_layers[[i]]

    base_scale <- tensor_to_numeric(layer$base_scale)
    fourier_scale <- tensor_to_numeric(layer$fourier_scale)

    out[[i]] <- data.frame(
      layer = i,
      feature = seq_along(base_scale),
      base_scale = base_scale,
      fourier_scale = fourier_scale,
      fourier_to_base_ratio = abs(fourier_scale) / (abs(base_scale) + 1e-8)
    )
  }

  do.call(rbind, out)
}


#' Extract KAF Fourier parameters
#'
#' Extracts the trainable Random Fourier Feature weights and biases from a KAF
#' model.
#'
#' @param object A fitted object returned by `kaf_fit()`, or a raw `nn_kaf`
#'   torch module.
#' @param layer Optional integer. If supplied, only extract parameters from this
#'   layer.
#'
#' @return A data frame with Fourier weights and biases.
#'
#' @export
extract_fourier_params <- function(object, layer = NULL) {
  model <- if (inherits(object, "kaf_fit")) object$model else object

  if (is.null(model$kaf_layers) || is.null(model$num_layers)) {
    stop("`object` must be a fitted KAF object or an `nn_kaf` model.",
         call. = FALSE)
  }

  layers <- if (is.null(layer)) {
    seq_len(model$num_layers)
  } else {
    if (!is.numeric(layer) || length(layer) != 1 ||
        layer < 1 || layer > model$num_layers) {
      stop("`layer` must be an integer between 1 and the number of KAF layers.",
           call. = FALSE)
    }

    as.integer(layer)
  }

  out <- vector("list", length(layers))

  for (j in seq_along(layers)) {
    i <- layers[[j]]
    kaf_layer <- model$kaf_layers[[i]]

    weight <- as.array(kaf_layer$rff$weight$detach())
    bias <- tensor_to_numeric(kaf_layer$rff$bias)

    grid <- expand.grid(
      input_feature = seq_len(nrow(weight)),
      grid = seq_len(ncol(weight))
    )

    grid$layer <- i
    grid$weight <- as.vector(weight)
    grid$bias <- bias[grid$grid]

    out[[j]] <- grid[, c(
      "layer",
      "input_feature",
      "grid",
      "weight",
      "bias"
    )]
  }

  do.call(rbind, out)
}


#' Plot KAF branch scales
#'
#' Visualizes the learned base/GELU and Fourier branch scales for a selected KAF
#' layer.
#'
#' @param object A fitted object returned by `kaf_fit()`, or a raw `nn_kaf`
#'   torch module.
#' @param layer Integer. Layer to inspect.
#' @param type Character. Either `"ratio"` or `"branch"`.
#' @param ... Additional arguments passed to base plotting functions.
#'
#' @return Invisibly returns the extracted scale data.
#'
#' @export
plot_kaf_scales <- function(object,
                            layer = 1,
                            type = c("ratio", "branch"),
                            ...) {
  type <- match.arg(type)

  scales <- extract_kaf_scales(object)

  if (!layer %in% scales$layer) {
    stop("`layer` is outside the available layer range.", call. = FALSE)
  }

  layer_scales <- scales[scales$layer == layer, ]

  if (type == "ratio") {
    graphics::plot(
      layer_scales$feature,
      layer_scales$fourier_to_base_ratio,
      type = "h",
      xlab = "Feature",
      ylab = "|Fourier scale| / |Base scale|",
      main = paste("KAF Fourier-to-base ratio, layer", layer),
      ...
    )

    graphics::points(
      layer_scales$feature,
      layer_scales$fourier_to_base_ratio
    )
  } else {
    graphics::matplot(
      layer_scales$feature,
      cbind(layer_scales$base_scale, layer_scales$fourier_scale),
      type = "l",
      lty = 1,
      lwd = 2,
      xlab = "Feature",
      ylab = "Scale",
      main = paste("KAF branch scales, layer", layer),
      ...
    )

    graphics::legend(
      "topright",
      legend = c("Base/GELU scale", "Fourier scale"),
      lty = 1,
      lwd = 2,
      bty = "n"
    )
  }

  invisible(layer_scales)
}

test_that("extract_kaf_scales() returns expected columns", {
  skip_if_no_torch()
  torch::torch_manual_seed(123)

  model <- nn_kaf(
    layers = c(3, 8, 1),
    num_grids = 4
  )

  scales <- extract_kaf_scales(model)

  expect_s3_class(scales, "data.frame")
  expect_true(all(c(
    "layer",
    "feature",
    "base_scale",
    "fourier_scale",
    "fourier_to_base_ratio"
  ) %in% names(scales)))

  expect_equal(nrow(scales), 3 + 8)
})


test_that("extract_fourier_params() returns expected columns", {
  skip_if_no_torch()
  torch::torch_manual_seed(123)

  model <- nn_kaf(
    layers = c(3, 8, 1),
    num_grids = 4
  )

  params <- extract_fourier_params(model, layer = 1)

  expect_s3_class(params, "data.frame")
  expect_true(all(c(
    "layer",
    "input_feature",
    "grid",
    "weight",
    "bias"
  ) %in% names(params)))

  expect_equal(nrow(params), 3 * 4)
  expect_true(all(params$layer == 1))
})


test_that("extract_fourier_params() rejects invalid layer", {
  skip_if_no_torch()
  torch::torch_manual_seed(123)

  model <- nn_kaf(
    layers = c(3, 8, 1),
    num_grids = 4
  )

  expect_error(
    extract_fourier_params(model, layer = 99),
    "`layer` must be an integer between 1 and the number of KAF layers"
  )
})


test_that("plot_kaf_scales() invisibly returns scale data", {
  skip_if_no_torch()
  torch::torch_manual_seed(123)

  model <- nn_kaf(
    layers = c(3, 8, 1),
    num_grids = 4
  )

  result <- plot_kaf_scales(model, layer = 1, type = "ratio")

  expect_s3_class(result, "data.frame")
  expect_true(all(result$layer == 1))
})

#' Print a fitted KAF model
#'
#' @param x A fitted object returned by `kaf_fit()`.
#' @param ... Unused.
#'
#' @return Invisibly returns `x`.
#'
#' @method print kaf_fit
#' @export
print.kaf_fit <- function(x, ...) {
  architecture <- c(x$input_dim, x$hidden, x$output_dim)

  cat("<kaf_fit>\n")
  cat("Task:         ", x$task, "\n", sep = "")
  if (!is.null(x$class_levels)) {
    cat("Classes:      ", paste(x$class_levels, collapse = ", "), "\n", sep = "")
  }
  if (!is.null(x$formula)) {
    cat("Formula:      ", deparse(x$formula), "\n", sep = "")
  }
  cat("Architecture: ", paste(architecture, collapse = " -> "), "\n", sep = "")
  cat("Fourier grids:", x$num_grids, "\n")
  cat("Epochs:       ", length(x$loss_history), "\n", sep = "")
  if (!is.null(x$batch_size)) {
    cat("Batch size:   ", x$batch_size, "\n", sep = "")
  }

  if (!is.null(x$validation_loss_history) &&
      any(!is.na(x$validation_loss_history))) {
    cat("Validation:   yes\n")
  } else {
    cat("Validation:   no\n")
  }

  cat("Standardize x:", if (isTRUE(x$standardize_x)) "yes" else "no", "\n")
  cat("Standardize y:", if (isTRUE(x$standardize_y)) "yes" else "no", "\n")

  if (!is.null(x$stopped_epoch) && !is.na(x$stopped_epoch)) {
    cat("Stopped at:   ", x$stopped_epoch, "\n", sep = "")
  }
  final_train_loss <- x$train_loss_history[[length(x$train_loss_history)]]

  cat(
    "Final train loss: ",
    format(final_train_loss, digits = 6),
    "\n",
    sep = ""
  )

  has_validation <- !is.null(x$validation_loss_history) &&
    any(!is.na(x$validation_loss_history))

  if (has_validation) {
    final_val_loss <- utils::tail(stats::na.omit(x$validation_loss_history), 1)

    cat(
      "Final val loss:   ",
      format(final_val_loss, digits = 6),
      "\n",
      sep = ""
    )
  }

  if (!is.null(x$best_loss)) {
    best_label <- if (has_validation) {
      "Best val loss:    "
    } else {
      "Best train loss:  "
    }

    cat(
      best_label,
      format(x$best_loss, digits = 6),
      " at epoch ",
      x$best_epoch,
      "\n",
      sep = ""
    )
  }

  invisible(x)
}


#' Plot a fitted KAF model
#'
#' @param x A fitted object returned by `kaf_fit()`.
#' @param type Character. Either `"loss"` or `"fit"`.
#' @param newdata Optional predictors used when `type = "fit"`.
#' @param y Optional observed target values used when `type = "fit"`.
#' @param ... Additional arguments passed to base plotting functions.
#'
#' @return Invisibly returns `x`.
#'
#' @method plot kaf_fit
#' @export
plot.kaf_fit <- function(x,
                         type = c("loss", "fit"),
                         newdata = NULL,
                         y = NULL,
                         ...) {
  type <- match.arg(type)

  if (type == "loss") {
    graphics::plot(
      seq_along(x$train_loss_history),
      x$train_loss_history,
      type = "l",
      xlab = "Epoch",
      ylab = if (isTRUE(x$standardize_y)) "MSE on standardized target" else "MSE",
      main = "KAF training loss",
      ...
    )

    if (!is.null(x$validation_loss_history) &&
        any(!is.na(x$validation_loss_history))) {
      graphics::lines(
        seq_along(x$validation_loss_history),
        x$validation_loss_history,
        lty = 2,
        lwd = 2
      )

      graphics::legend(
        "topright",
        legend = c("Training", "Validation"),
        lty = c(1, 2),
        lwd = c(1, 2),
        bty = "n"
      )
    }

    return(invisible(x))
  }

  if (is.null(newdata) || is.null(y)) {
    stop("`newdata` and `y` are required when `type = 'fit'`.",
         call. = FALSE)
  }

  if (x$output_dim != 1) {
    stop("`plot(..., type = 'fit')` currently supports single-output models only.",
         call. = FALSE)
  }

  x_tensor <- as_float_tensor_matrix(newdata, "newdata")
  pred <- stats::predict(x, newdata)

  y_tensor <- as_float_tensor_target(y, "y")
  y_vec <- as.numeric(y_tensor)

  if (x_tensor$shape[[2]] == 1) {
    x_vec <- as.numeric(x_tensor)
    ord <- order(x_vec)

    graphics::plot(
      x_vec[ord],
      y_vec[ord],
      type = "l",
      lwd = 2,
      xlab = "x",
      ylab = "y",
      main = "KAF fitted function",
      ...
    )

    graphics::lines(
      x_vec[ord],
      pred[ord],
      lwd = 2,
      lty = 2
    )

    graphics::legend(
      "topright",
      legend = c("Observed", "Predicted"),
      lty = c(1, 2),
      lwd = 2,
      bty = "n"
    )
  } else {
    graphics::plot(
      y_vec,
      pred,
      xlab = "Observed",
      ylab = "Predicted",
      main = "KAF observed vs predicted",
      ...
    )

    graphics::abline(0, 1, lty = 2)
  }

  invisible(x)
}

#' rkaf: Kolmogorov-Arnold Fourier Networks in R
#'
#' `rkaf` provides an R implementation of Kolmogorov-Arnold Fourier Networks
#' using the `torch` backend.
#'
#' The package supports:
#'
#' - regression
#' - binary classification
#' - multiclass classification
#' - formula and matrix interfaces
#' - mini-batch training
#' - validation splits
#' - early stopping
#' - automatic standardization
#' - best-model restoration
#' - KAF-specific diagnostics
#'
#' @keywords internal
"_PACKAGE"

as_float_tensor_matrix <- function(x, name = "x") {
  if (inherits(x, "torch_tensor")) {
    if (length(x$shape) != 2) {
      stop("`", name, "` must be a 2D tensor.", call. = FALSE)
    }
    return(x$to(dtype = torch::torch_float()))
  }

  if (is.data.frame(x)) {
    x <- stats::model.matrix(~ . - 1, data = x)
  }

  if (is.vector(x)) {
    x <- matrix(x, ncol = 1)
  }

  if (!is.matrix(x)) {
    stop("`", name, "` must be a matrix, data frame, vector, or torch tensor.",
         call. = FALSE)
  }

  storage.mode(x) <- "double"

  torch::torch_tensor(x, dtype = torch::torch_float())
}


as_float_tensor_target <- function(y, name = "y") {
  if (inherits(y, "torch_tensor")) {
    if (length(y$shape) == 1) {
      y <- y$unsqueeze(2)
    }
    if (length(y$shape) != 2) {
      stop("`", name, "` must be a vector, matrix, or 2D tensor.",
           call. = FALSE)
    }
    return(y$to(dtype = torch::torch_float()))
  }

  if (is.data.frame(y)) {
    y <- as.matrix(y)
  }

  if (is.vector(y)) {
    y <- matrix(y, ncol = 1)
  }

  if (!is.matrix(y)) {
    stop("`", name, "` must be a vector, matrix, data frame, or torch tensor.",
         call. = FALSE)
  }

  storage.mode(y) <- "double"

  torch::torch_tensor(y, dtype = torch::torch_float())
}

clone_state_dict <- function(state_dict) {
  lapply(state_dict, function(x) {
    x$detach()$clone()
  })
}

compute_tensor_standardizer <- function(x) {
  mat <- as.matrix(as.array(x))

  center <- colMeans(mat)
  scale <- apply(mat, 2, stats::sd)

  scale[is.na(scale) | scale == 0] <- 1

  list(
    center = center,
    scale = scale
  )
}


apply_tensor_standardizer <- function(x, standardizer) {
  mat <- as.matrix(as.array(x))

  mat <- sweep(mat, 2, standardizer$center, FUN = "-")
  mat <- sweep(mat, 2, standardizer$scale, FUN = "/")

  torch::torch_tensor(mat, dtype = torch::torch_float())
}


invert_tensor_standardizer_array <- function(x, standardizer) {
  mat <- as.matrix(x)

  mat <- sweep(mat, 2, standardizer$scale, FUN = "*")
  mat <- sweep(mat, 2, standardizer$center, FUN = "+")

  mat
}

infer_kaf_task <- function(y, task = c("auto", "regression", "binary", "multiclass")) {
  task <- match.arg(task)

  if (task != "auto") {
    return(task)
  }

  if (is.factor(y) || is.character(y) || is.logical(y)) {
    n_classes <- length(unique(stats::na.omit(y)))

    if (n_classes == 2) {
      return("binary")
    }

    if (n_classes > 2) {
      return("multiclass")
    }

    stop("Classification targets must contain at least two classes.",
         call. = FALSE)
  }

  "regression"
}


prepare_kaf_target <- function(y,
                               task = c("regression", "binary", "multiclass"),
                               name = "y",
                               class_levels = NULL) {
  task <- match.arg(task)

  if (task == "regression") {
    return(list(
      tensor = as_float_tensor_target(y, name),
      output_dim = NULL,
      class_levels = NULL
    ))
  }

  if (inherits(y, "torch_tensor")) {
    y_vec <- as.array(y)
  } else if (is.data.frame(y)) {
    if (ncol(y) != 1) {
      stop("Classification target `", name, "` must have one column.",
           call. = FALSE)
    }
    y_vec <- y[[1]]
  } else {
    y_vec <- y
  }

  if (is.matrix(y_vec)) {
    if (ncol(y_vec) != 1) {
      stop("Classification target `", name, "` must have one column.",
           call. = FALSE)
    }
    y_vec <- as.vector(y_vec)
  }

  if (is.null(class_levels)) {
    if (is.factor(y_vec)) {
      class_levels <- levels(y_vec)
    } else {
      class_levels <- sort(unique(stats::na.omit(as.character(y_vec))))
    }
  }

  y_chr <- as.character(y_vec)
  y_idx <- match(y_chr, class_levels)

  if (any(is.na(y_idx))) {
    stop("`", name, "` contains class values not seen during training.",
         call. = FALSE)
  }

  if (task == "binary") {
    if (length(class_levels) != 2) {
      stop("Binary classification requires exactly two classes.",
           call. = FALSE)
    }

    # Negative class = class_levels[1], positive class = class_levels[2].
    y01 <- as.numeric(y_idx == 2L)

    return(list(
      tensor = torch::torch_tensor(matrix(y01, ncol = 1), dtype = torch::torch_float()),
      output_dim = 1L,
      class_levels = class_levels
    ))
  }

  if (length(class_levels) < 3) {
    stop("Multiclass classification requires at least three classes.",
         call. = FALSE)
  }

  # R torch cross-entropy expects 1-based class indices.
  # Class 1 = class_levels[1], class 2 = class_levels[2], etc.
  y_class <- as.integer(y_idx)

  list(
    tensor = torch::torch_tensor(y_class, dtype = torch::torch_long()),
    output_dim = length(class_levels),
    class_levels = class_levels
  )
}


kaf_loss <- function(pred, target, task) {
  if (task == "regression") {
    return(torch::nnf_mse_loss(pred, target))
  }

  if (task == "binary") {
    return(torch::nnf_binary_cross_entropy_with_logits(pred, target))
  }

  if (task == "multiclass") {
    return(torch::nnf_cross_entropy(pred, target))
  }

  stop("Unsupported task.", call. = FALSE)
}


kaf_prediction_to_r <- function(pred,
                                object,
                                type = c("response", "prob", "class", "link"),
                                threshold = 0.5,
                                as_tensor = FALSE) {
  type <- match.arg(type)

  task <- object$task

  if (task == "regression") {
    if (type %in% c("prob", "class")) {
      stop("`type = 'prob'` and `type = 'class'` are only available for classification.",
           call. = FALSE)
    }

    pred_r <- as.array(pred)

    if (isTRUE(object$standardize_y) && !is.null(object$y_standardizer)) {
      pred_r <- invert_tensor_standardizer_array(pred_r, object$y_standardizer)
    }

    if (isTRUE(as_tensor)) {
      return(torch::torch_tensor(pred_r, dtype = torch::torch_float()))
    }

    if (object$output_dim == 1) {
      return(as.numeric(pred_r))
    }

    return(pred_r)
  }

  if (task == "binary") {
    if (type == "link") {
      logits <- as.array(pred)

      if (isTRUE(as_tensor)) {
        return(pred)
      }

      return(as.numeric(logits))
    }

    probs_tensor <- torch::torch_sigmoid(pred)
    probs <- as.numeric(as.array(probs_tensor))

    if (type %in% c("response", "prob")) {
      if (isTRUE(as_tensor)) {
        return(probs_tensor)
      }

      return(probs)
    }

    if (isTRUE(as_tensor)) {
      stop("`as_tensor = TRUE` is not supported with `type = 'class'`.",
           call. = FALSE)
    }

    classes <- ifelse(
      probs >= threshold,
      object$class_levels[[2]],
      object$class_levels[[1]]
    )

    return(factor(classes, levels = object$class_levels))
  }

  if (task == "multiclass") {
    if (type == "link") {
      logits <- as.array(pred)
      colnames(logits) <- object$class_levels

      if (isTRUE(as_tensor)) {
        return(pred)
      }

      return(logits)
    }

    probs_tensor <- torch::nnf_softmax(pred, dim = 2)
    probs <- as.array(probs_tensor)
    colnames(probs) <- object$class_levels

    if (type %in% c("response", "prob")) {
      if (isTRUE(as_tensor)) {
        return(probs_tensor)
      }

      return(probs)
    }

    if (isTRUE(as_tensor)) {
      stop("`as_tensor = TRUE` is not supported with `type = 'class'`.",
           call. = FALSE)
    }

    class_id <- max.col(probs)
    return(factor(object$class_levels[class_id], levels = object$class_levels))
  }

  stop("Unsupported task.", call. = FALSE)
}


slice_kaf_target <- function(y_tensor, idx, task) {
  if (task == "multiclass") {
    return(y_tensor[idx])
  }

  y_tensor[idx, , drop = FALSE]
}

#' Fit a Kolmogorov-Arnold Fourier Network
#'
#' Fits a KAF model for regression using mean squared error loss.
#'
#' @param x Matrix, data frame, vector, or 2D torch tensor of predictors.
#' @param y Vector, matrix, data frame, or torch tensor of targets.
#' @param task Character. One of `"auto"`, `"regression"`, `"binary"`, or
#'   `"multiclass"`. With `"auto"`, factor, character, and logical targets are
#'   treated as classification; numeric targets are treated as regression.
#' @param hidden Integer vector. Hidden layer sizes.
#' @param num_grids Integer. Number of Fourier frequencies per KAF layer.
#' @param dropout Numeric. Dropout probability.
#' @param use_layernorm Logical. Whether to apply layer normalization.
#' @param fourier_init_scale Numeric. Initial scale of the Fourier branch.
#' @param epochs Integer. Maximum number of training epochs.
#' @param lr Numeric. Learning rate.
#' @param batch_size Optional integer. Mini-batch size. If `NULL`, full-batch
#'   training is used.
#' @param shuffle Logical. Whether to shuffle training rows each epoch.
#' @param validation_split Numeric in `[0, 1)`. Fraction of rows to reserve for
#'   validation. Ignored if `x_val` and `y_val` are supplied.
#' @param x_val Optional validation predictors.
#' @param y_val Optional validation targets.
#' @param weight_decay Numeric. Adam weight decay.
#' @param standardize_x Logical. Whether to standardize predictors using the
#'   training-set mean and standard deviation.
#' @param standardize_y Logical or `NULL`. Whether to standardize regression
#'   targets using the training-set mean and standard deviation. If `NULL`,
#'   targets are standardized for regression and not standardized for
#'   classification. Predictions are automatically transformed back to the
#'   original target scale.
#' @param patience Optional integer. Number of epochs without improvement before
#'   early stopping.
#' @param verbose Logical. Whether to print progress.
#' @param print_every Integer. Print frequency.
#' @param seed Optional integer random seed.
#' @param restore_best Logical. Whether to restore the best observed model state
#'   after training.
#' @param min_delta Numeric. Minimum loss improvement required to update the
#'   best model state.
#'
#' @return An object of class `"kaf_fit"`.
#'
#' @export
kaf_fit <- function(x,
                    y,
                    task = c("auto", "regression", "binary", "multiclass"),
                    hidden = c(64, 64),
                    num_grids = 16,
                    dropout = 0,
                    use_layernorm = TRUE,
                    fourier_init_scale = 1e-2,
                    epochs = 1000,
                    lr = 1e-3,
                    batch_size = NULL,
                    shuffle = TRUE,
                    validation_split = 0,
                    x_val = NULL,
                    y_val = NULL,
                    weight_decay = 0,
                    standardize_x = TRUE,
                    standardize_y = NULL,
                    patience = NULL,
                    verbose = TRUE,
                    print_every = 100,
                    seed = NULL,
                    restore_best = TRUE,
                    min_delta = 0) {
  if (!is.null(seed)) {
    torch::torch_manual_seed(seed)
    set.seed(seed)
  }

  if (!is.numeric(epochs) || length(epochs) != 1 || epochs < 1) {
    stop("`epochs` must be a positive integer.", call. = FALSE)
  }

  if (!is.null(batch_size) &&
      (!is.numeric(batch_size) || length(batch_size) != 1 || batch_size < 1)) {
    stop("`batch_size` must be `NULL` or a positive integer.", call. = FALSE)
  }

  if (!is.numeric(validation_split) ||
      length(validation_split) != 1 ||
      validation_split < 0 ||
      validation_split >= 1) {
    stop("`validation_split` must be in [0, 1).", call. = FALSE)
  }

  if (!is.null(patience) &&
      (!is.numeric(patience) || length(patience) != 1 || patience < 1)) {
    stop("`patience` must be `NULL` or a positive integer.", call. = FALSE)
  }

  task <- infer_kaf_task(y, task)

  if (is.null(standardize_y)) {
    standardize_y <- task == "regression"
  }

  if (task != "regression" && isTRUE(standardize_y)) {
    stop("`standardize_y = TRUE` is only supported for regression.",
         call. = FALSE)
  }

  x_tensor <- as_float_tensor_matrix(x, "x")

  target_info <- prepare_kaf_target(y, task = task, name = "y")
  y_tensor <- target_info$tensor
  class_levels <- target_info$class_levels

  if (x_tensor$shape[[1]] != y_tensor$shape[[1]]) {
    stop("`x` and `y` must have the same number of rows.", call. = FALSE)
  }

  n <- x_tensor$shape[[1]]

  has_explicit_validation <- !is.null(x_val) || !is.null(y_val)

  if (has_explicit_validation && (is.null(x_val) || is.null(y_val))) {
    stop("Both `x_val` and `y_val` must be supplied for explicit validation.",
         call. = FALSE)
  }

  if (has_explicit_validation) {
    x_val_tensor <- as_float_tensor_matrix(x_val, "x_val")

    y_val_info <- prepare_kaf_target(
      y_val,
      task = task,
      name = "y_val",
      class_levels = class_levels
    )

    y_val_tensor <- y_val_info$tensor

    if (x_val_tensor$shape[[1]] != y_val_tensor$shape[[1]]) {
      stop("`x_val` and `y_val` must have the same number of rows.",
           call. = FALSE)
    }

    if (x_val_tensor$shape[[2]] != x_tensor$shape[[2]]) {
      stop("`x_val` must have the same number of columns as `x`.",
           call. = FALSE)
    }
  } else if (validation_split > 0) {
    n_val <- max(1L, floor(n * validation_split))

    if (n_val >= n) {
      stop("`validation_split` leaves no rows for training.", call. = FALSE)
    }

    val_idx <- sample.int(n, n_val)
    train_idx <- setdiff(seq_len(n), val_idx)

    x_val_tensor <- x_tensor[val_idx, , drop = FALSE]
    y_val_tensor <- slice_kaf_target(y_tensor, val_idx, task)

    x_tensor <- x_tensor[train_idx, , drop = FALSE]
    y_tensor <- slice_kaf_target(y_tensor, train_idx, task)
  } else {
    x_val_tensor <- NULL
    y_val_tensor <- NULL
  }

  if (task %in% c("regression", "binary") &&
      !is.null(y_val_tensor) &&
      y_val_tensor$shape[[2]] != y_tensor$shape[[2]]) {
    stop("`y_val` must have the same number of columns as `y`.",
         call. = FALSE)
  }

  x_standardizer <- NULL
  y_standardizer <- NULL

  if (isTRUE(standardize_x)) {
    x_standardizer <- compute_tensor_standardizer(x_tensor)

    x_tensor <- apply_tensor_standardizer(x_tensor, x_standardizer)

    if (!is.null(x_val_tensor)) {
      x_val_tensor <- apply_tensor_standardizer(x_val_tensor, x_standardizer)
    }
  }

  if (isTRUE(standardize_y)) {
    y_standardizer <- compute_tensor_standardizer(y_tensor)

    y_tensor <- apply_tensor_standardizer(y_tensor, y_standardizer)

    if (!is.null(y_val_tensor)) {
      y_val_tensor <- apply_tensor_standardizer(y_val_tensor, y_standardizer)
    }
  }

  input_dim <- x_tensor$shape[[2]]

  output_dim <- if (task == "multiclass") {
    length(class_levels)
  } else if (task == "binary") {
    1L
  } else {
    y_tensor$shape[[2]]
  }

  model <- kaf(
    input_dim = input_dim,
    output_dim = output_dim,
    hidden = hidden,
    num_grids = num_grids,
    dropout = dropout,
    use_layernorm = use_layernorm,
    fourier_init_scale = fourier_init_scale
  )

  optimizer <- torch::optim_adam(
    model$parameters,
    lr = lr,
    weight_decay = weight_decay
  )

  train_loss_history <- numeric(epochs)
  validation_loss_history <- rep(NA_real_, epochs)

  best_loss <- Inf
  best_epoch <- NA_integer_
  best_state <- NULL
  epochs_without_improvement <- 0L
  stopped_epoch <- NA_integer_

  n_train <- x_tensor$shape[[1]]

  if (is.null(batch_size) || batch_size >= n_train) {
    batch_size <- n_train
  } else {
    batch_size <- as.integer(batch_size)
  }

  evaluate_loss <- function(x_eval, y_eval) {
    model$eval()

    loss <- torch::with_no_grad({
      pred <- model(x_eval)
      kaf_loss(pred, y_eval, task)
    })

    loss$item()
  }

  for (epoch in seq_len(epochs)) {
    model$train()

    row_order <- if (isTRUE(shuffle)) {
      sample.int(n_train)
    } else {
      seq_len(n_train)
    }

    batch_losses <- numeric(ceiling(n_train / batch_size))
    batch_counter <- 0L

    for (start in seq(1L, n_train, by = batch_size)) {
      end <- min(start + batch_size - 1L, n_train)
      idx <- row_order[start:end]

      xb <- x_tensor[idx, , drop = FALSE]
      yb <- slice_kaf_target(y_tensor, idx, task)

      optimizer$zero_grad()

      pred <- model(xb)
      loss <- kaf_loss(pred, yb, task)

      loss$backward()
      optimizer$step()

      batch_counter <- batch_counter + 1L
      batch_losses[[batch_counter]] <- loss$item()
    }

    train_loss <- mean(batch_losses)
    train_loss_history[[epoch]] <- train_loss

    validation_loss <- NA_real_

    if (!is.null(x_val_tensor)) {
      validation_loss <- evaluate_loss(x_val_tensor, y_val_tensor)
      validation_loss_history[[epoch]] <- validation_loss
    }

    monitored_loss <- if (!is.na(validation_loss)) {
      validation_loss
    } else {
      train_loss
    }

    if (monitored_loss < best_loss - min_delta) {
      best_loss <- monitored_loss
      best_epoch <- epoch
      best_state <- clone_state_dict(model$state_dict())
      epochs_without_improvement <- 0L
    } else {
      epochs_without_improvement <- epochs_without_improvement + 1L
    }

    if (isTRUE(verbose) && epoch %% print_every == 0) {
      if (!is.na(validation_loss)) {
        cat(sprintf(
          "Epoch %4d | train MSE: %.6f | val MSE: %.6f\n",
          epoch,
          train_loss,
          validation_loss
        ))
      } else {
        cat(sprintf(
          "Epoch %4d | train MSE: %.6f\n",
          epoch,
          train_loss
        ))
      }
    }

    if (!is.null(patience) && epochs_without_improvement >= patience) {
      stopped_epoch <- epoch

      if (isTRUE(verbose)) {
        cat(sprintf(
          "Early stopping at epoch %d. Best epoch: %d | best loss: %.6f\n",
          stopped_epoch,
          best_epoch,
          best_loss
        ))
      }

      break
    }
  }

  if (isTRUE(restore_best) && !is.null(best_state)) {
    model$load_state_dict(best_state)
  }

  model$eval()

  completed_epochs <- if (is.na(stopped_epoch)) epochs else stopped_epoch

  train_loss_history <- train_loss_history[seq_len(completed_epochs)]
  validation_loss_history <- validation_loss_history[seq_len(completed_epochs)]

  out <- list(
    model = model,
    loss_history = train_loss_history,
    train_loss_history = train_loss_history,
    validation_loss_history = validation_loss_history,
    best_loss = best_loss,
    best_epoch = best_epoch,
    restore_best = restore_best,
    stopped_epoch = stopped_epoch,
    completed_epochs = completed_epochs,
    input_dim = input_dim,
    output_dim = output_dim,
    hidden = hidden,
    num_grids = num_grids,
    batch_size = batch_size,
    validation_split = validation_split,
    weight_decay = weight_decay,
    x_standardizer = x_standardizer,
    y_standardizer = y_standardizer,
    standardize_x = standardize_x,
    standardize_y = standardize_y,
    class_levels = class_levels,
    task = task
  )

  class(out) <- "kaf_fit"
  out
}


#' Predict from a fitted KAF model
#'
#' @param object A fitted object returned by `kaf_fit()`.
#' @param newdata Matrix, data frame, vector, or 2D torch tensor.
#' @param type Character. Prediction type. `"response"` returns regression
#'   predictions for regression, probabilities for classification. `"prob"`
#'   returns probabilities for classification. `"class"` returns predicted
#'   classes. `"link"` returns raw model logits/outputs.
#' @param threshold Numeric. Classification threshold used for binary class
#'   predictions.
#' @param as_tensor Logical. If `TRUE`, return a torch tensor where supported.
#' @param ... Unused.
#'
#' @return Predictions as a vector, matrix, factor, or torch tensor.
#'
#' @method predict kaf_fit
#' @export
predict.kaf_fit <- function(object,
                            newdata,
                            type = c("response", "prob", "class", "link"),
                            threshold = 0.5,
                            as_tensor = FALSE,
                            ...) {
  if (!inherits(object, "kaf_fit")) {
    stop("`object` must inherit from class 'kaf_fit'.", call. = FALSE)
  }

  type <- match.arg(type)

  newdata_processed <- if (!is.null(object$formula) && is.data.frame(newdata)) {
    model_matrix_from_kaf_formula(object, newdata)
  } else {
    newdata
  }

  x_tensor <- as_float_tensor_matrix(newdata_processed, "newdata")

  if (x_tensor$shape[[2]] != object$input_dim) {
    stop(
      "`newdata` must have ",
      object$input_dim,
      " columns/features.",
      call. = FALSE
    )
  }

  if (isTRUE(object$standardize_x) && !is.null(object$x_standardizer)) {
    x_tensor <- apply_tensor_standardizer(x_tensor, object$x_standardizer)
  }

  object$model$eval()

  pred <- torch::with_no_grad({
    object$model(x_tensor)
  })

  kaf_prediction_to_r(
    pred = pred,
    object = object,
    type = type,
    threshold = threshold,
    as_tensor = as_tensor
  )
}

test_that("kaf() creates a model with the expected output shape", {
  skip_if_no_torch()
  torch::torch_manual_seed(123)

  model <- kaf(
    input_dim = 3,
    output_dim = 1,
    hidden = c(8, 8),
    num_grids = 4
  )

  x <- torch::torch_randn(5, 3)
  y <- model(x)

  expect_equal(as.integer(y$shape), c(5L, 1L))
})


test_that("kaf_fit() returns a fitted KAF object", {
  skip_if_no_torch()
  torch::torch_manual_seed(123)

  x <- matrix(seq(-1, 1, length.out = 32), ncol = 1)
  y <- sin(2 * pi * x)

  fit <- kaf_fit(
    x = x,
    y = y,
    hidden = c(8),
    num_grids = 4,
    use_layernorm = FALSE,
    epochs = 5,
    lr = 1e-3,
    verbose = FALSE,
    seed = 123
  )

  expect_s3_class(fit, "kaf_fit")
  expect_equal(length(fit$loss_history), 5L)
  expect_equal(fit$input_dim, 1)
  expect_equal(fit$output_dim, 1)
})


test_that("predict.kaf_fit() returns numeric predictions", {
  skip_if_no_torch()
  torch::torch_manual_seed(123)

  x <- matrix(seq(-1, 1, length.out = 32), ncol = 1)
  y <- sin(2 * pi * x)

  fit <- kaf_fit(
    x = x,
    y = y,
    hidden = c(8),
    num_grids = 4,
    use_layernorm = FALSE,
    epochs = 5,
    lr = 1e-3,
    verbose = FALSE,
    seed = 123
  )

  pred <- predict(fit, x)

  expect_type(pred, "double")
  expect_equal(length(pred), nrow(x))
})


test_that("predict.kaf_fit() can return a tensor", {
  skip_if_no_torch()
  torch::torch_manual_seed(123)

  x <- matrix(seq(-1, 1, length.out = 32), ncol = 1)
  y <- sin(2 * pi * x)

  fit <- kaf_fit(
    x = x,
    y = y,
    hidden = c(8),
    num_grids = 4,
    use_layernorm = FALSE,
    epochs = 5,
    lr = 1e-3,
    verbose = FALSE,
    seed = 123
  )

  pred <- predict(fit, x, as_tensor = TRUE)

  expect_true(inherits(pred, "torch_tensor"))
  expect_equal(as.integer(pred$shape), c(32L, 1L))
})


test_that("predict.kaf_fit() rejects wrong feature dimensions", {
  skip_if_no_torch()
  torch::torch_manual_seed(123)

  x <- matrix(rnorm(40), ncol = 2)
  y <- rowSums(x)

  fit <- kaf_fit(
    x = x,
    y = y,
    hidden = c(8),
    num_grids = 4,
    epochs = 5,
    verbose = FALSE,
    seed = 123
  )

  bad_x <- matrix(rnorm(30), ncol = 3)

  expect_error(
    predict(fit, bad_x),
    "`newdata` must have 2 columns/features"
  )
})

test_that("kaf_fit() tracks best loss and best epoch", {
  skip_if_no_torch()
  torch::torch_manual_seed(123)

  x <- matrix(seq(-1, 1, length.out = 32), ncol = 1)
  y <- sin(2 * pi * x)

  fit <- kaf_fit(
    x = x,
    y = y,
    hidden = c(8),
    num_grids = 4,
    use_layernorm = FALSE,
    epochs = 10,
    lr = 1e-3,
    verbose = FALSE,
    seed = 123,
    restore_best = TRUE
  )

  expect_s3_class(fit, "kaf_fit")
  expect_true(is.numeric(fit$best_loss))
  expect_true(is.numeric(fit$best_epoch))
  expect_true(fit$best_epoch >= 1)
  expect_true(fit$best_epoch <= 10)
  expect_equal(fit$best_loss, min(fit$loss_history))
  expect_true(isTRUE(fit$restore_best))
})

test_that("kaf_fit() supports mini-batch training", {
  skip_if_no_torch()
  torch::torch_manual_seed(123)

  x <- matrix(seq(-1, 1, length.out = 64), ncol = 1)
  y <- sin(2 * pi * x)

  fit <- kaf_fit(
    x = x,
    y = y,
    hidden = c(8),
    num_grids = 4,
    use_layernorm = FALSE,
    epochs = 5,
    batch_size = 16,
    lr = 1e-3,
    verbose = FALSE,
    seed = 123
  )

  expect_s3_class(fit, "kaf_fit")
  expect_equal(fit$batch_size, 16L)
  expect_equal(length(fit$train_loss_history), 5L)
  expect_equal(length(fit$loss_history), 5L)
})


test_that("kaf_fit() supports validation split", {
  skip_if_no_torch()
  torch::torch_manual_seed(123)

  x <- matrix(seq(-1, 1, length.out = 64), ncol = 1)
  y <- sin(2 * pi * x)

  fit <- kaf_fit(
    x = x,
    y = y,
    hidden = c(8),
    num_grids = 4,
    use_layernorm = FALSE,
    epochs = 5,
    validation_split = 0.25,
    lr = 1e-3,
    verbose = FALSE,
    seed = 123
  )

  expect_s3_class(fit, "kaf_fit")
  expect_equal(length(fit$validation_loss_history), 5L)
  expect_true(any(!is.na(fit$validation_loss_history)))
  expect_equal(fit$validation_split, 0.25)
})


test_that("kaf_fit() supports explicit validation data", {
  skip_if_no_torch()
  torch::torch_manual_seed(123)

  x <- matrix(seq(-1, 1, length.out = 64), ncol = 1)
  y <- sin(2 * pi * x)

  x_train <- x[1:48, , drop = FALSE]
  y_train <- y[1:48, , drop = FALSE]

  x_val <- x[49:64, , drop = FALSE]
  y_val <- y[49:64, , drop = FALSE]

  fit <- kaf_fit(
    x = x_train,
    y = y_train,
    x_val = x_val,
    y_val = y_val,
    hidden = c(8),
    num_grids = 4,
    use_layernorm = FALSE,
    epochs = 5,
    lr = 1e-3,
    verbose = FALSE,
    seed = 123
  )

  expect_s3_class(fit, "kaf_fit")
  expect_true(any(!is.na(fit$validation_loss_history)))
})


test_that("kaf_fit() rejects incomplete explicit validation data", {
  skip_if_no_torch()
  x <- matrix(seq(-1, 1, length.out = 32), ncol = 1)
  y <- sin(2 * pi * x)

  expect_error(
    kaf_fit(
      x = x,
      y = y,
      x_val = x,
      hidden = c(8),
      epochs = 2,
      verbose = FALSE
    ),
    "Both `x_val` and `y_val` must be supplied"
  )
})


test_that("kaf_fit() supports early stopping", {
  skip_if_no_torch()
  torch::torch_manual_seed(123)

  x <- matrix(seq(-1, 1, length.out = 64), ncol = 1)
  y <- sin(2 * pi * x)

  fit <- kaf_fit(
    x = x,
    y = y,
    hidden = c(8),
    num_grids = 4,
    use_layernorm = FALSE,
    epochs = 50,
    lr = 1e-8,
    patience = 3,
    min_delta = 1,
    verbose = FALSE,
    seed = 123
  )

  expect_s3_class(fit, "kaf_fit")
  expect_true(!is.na(fit$stopped_epoch))
  expect_true(fit$completed_epochs < 50)
  expect_equal(length(fit$loss_history), fit$completed_epochs)
})

test_that("kaf_fit() stores predictor standardization parameters", {
  skip_if_no_torch()
  torch::torch_manual_seed(123)

  x <- cbind(
    a = seq(1, 100, length.out = 50),
    b = seq(-10, 10, length.out = 50)
  )
  y <- rowSums(x)

  fit <- kaf_fit(
    x = x,
    y = y,
    hidden = c(8),
    num_grids = 4,
    epochs = 5,
    verbose = FALSE,
    seed = 123,
    standardize_x = TRUE
  )

  expect_s3_class(fit, "kaf_fit")
  expect_true(isTRUE(fit$standardize_x))
  expect_false(is.null(fit$x_standardizer))
  expect_equal(length(fit$x_standardizer$center), 2)
  expect_equal(length(fit$x_standardizer$scale), 2)
})


test_that("kaf_fit() can disable predictor standardization", {
  skip_if_no_torch()
  torch::torch_manual_seed(123)

  x <- matrix(seq(-1, 1, length.out = 40), ncol = 1)
  y <- sin(2 * pi * x)

  fit <- kaf_fit(
    x = x,
    y = y,
    hidden = c(8),
    num_grids = 4,
    epochs = 5,
    verbose = FALSE,
    seed = 123,
    standardize_x = FALSE
  )

  expect_false(isTRUE(fit$standardize_x))
  expect_true(is.null(fit$x_standardizer))
})


test_that("kaf_fit() supports target standardization and inverse prediction", {
  skip_if_no_torch()
  torch::torch_manual_seed(123)

  x <- matrix(seq(-1, 1, length.out = 50), ncol = 1)
  y <- 1000 + 500 * sin(2 * pi * x)

  fit <- kaf_fit(
    x = x,
    y = y,
    hidden = c(8),
    num_grids = 4,
    use_layernorm = FALSE,
    epochs = 5,
    verbose = FALSE,
    seed = 123,
    standardize_x = TRUE,
    standardize_y = TRUE
  )

  pred <- predict(fit, x)

  expect_s3_class(fit, "kaf_fit")
  expect_true(isTRUE(fit$standardize_y))
  expect_false(is.null(fit$y_standardizer))
  expect_type(pred, "double")
  expect_equal(length(pred), nrow(x))

  # Predictions should be returned on the original target scale, not the
  # standardized scale.
  expect_true(mean(pred) > 100)
})


test_that("predict.kaf_fit() returns unstandardized tensor predictions when requested", {
  skip_if_no_torch()
  torch::torch_manual_seed(123)

  x <- matrix(seq(-1, 1, length.out = 50), ncol = 1)
  y <- 1000 + 500 * sin(2 * pi * x)

  fit <- kaf_fit(
    x = x,
    y = y,
    hidden = c(8),
    num_grids = 4,
    use_layernorm = FALSE,
    epochs = 5,
    verbose = FALSE,
    seed = 123,
    standardize_y = TRUE
  )

  pred <- predict(fit, x, as_tensor = TRUE)

  expect_true(inherits(pred, "torch_tensor"))
  expect_equal(as.integer(pred$shape), c(50L, 1L))
})

test_that("kaf_fit() standardizes regression targets by default", {
  skip_if_no_torch()
  torch::torch_manual_seed(123)

  x <- matrix(seq(-1, 1, length.out = 40), ncol = 1)
  y <- 100 + 50 * sin(2 * pi * x)

  fit <- kaf_fit(
    x = x,
    y = y,
    hidden = c(8),
    num_grids = 4,
    epochs = 5,
    verbose = FALSE,
    seed = 123
  )

  expect_s3_class(fit, "kaf_fit")
  expect_true(isTRUE(fit$standardize_y))
  expect_false(is.null(fit$y_standardizer))

  pred <- predict(fit, x)

  expect_type(pred, "double")
  expect_equal(length(pred), nrow(x))

  # Predictions should be returned on the original target scale.
  expect_true(mean(pred) > 10)
})

#' Kolmogorov-Arnold Fourier Layer
#'
#' Torch module implementing one KAF layer using a hybrid GELU and
#' trainable Random Fourier Features activation.
#'
#' The layer computes a feature-wise hybrid activation of the form:
#'
#' \deqn{
#' z = \alpha \odot GELU(x) + \beta \odot RFF(x)
#' }
#'
#' followed by a linear projection to the output dimension.
#'
#' @param input_dim Integer. Input dimension.
#' @param output_dim Integer. Output dimension.
#' @param num_grids Integer. Number of Fourier frequencies.
#' @param dropout Numeric. Dropout probability.
#' @param use_layernorm Logical. Whether to apply layer normalization before
#'   the Fourier feature block.
#' @param activation_expectation Numeric. Scaling constant for Fourier
#'   initialization.
#' @param fourier_init_scale Numeric. Initial scale of the Fourier component.
#'
#' @return A torch nn_module.
#'
#' @export
nn_kaf_layer <- torch::nn_module(
  "nn_kaf_layer",

  initialize = function(input_dim,
                        output_dim,
                        num_grids = 8,
                        dropout = 0,
                        use_layernorm = TRUE,
                        activation_expectation = 1.64,
                        fourier_init_scale = 1e-2) {
    if (!is.numeric(input_dim) || length(input_dim) != 1 || input_dim < 1) {
      stop("`input_dim` must be a positive integer.", call. = FALSE)
    }

    if (!is.numeric(output_dim) || length(output_dim) != 1 || output_dim < 1) {
      stop("`output_dim` must be a positive integer.", call. = FALSE)
    }

    if (!is.numeric(num_grids) || length(num_grids) != 1 || num_grids < 1) {
      stop("`num_grids` must be a positive integer.", call. = FALSE)
    }

    if (!is.numeric(dropout) || dropout < 0 || dropout >= 1) {
      stop("`dropout` must be in [0, 1).", call. = FALSE)
    }

    self$input_dim <- as.integer(input_dim)
    self$output_dim <- as.integer(output_dim)
    self$num_grids <- as.integer(num_grids)
    self$use_layernorm <- isTRUE(use_layernorm)

    if (self$use_layernorm) {
      self$layernorm <- torch::nn_layer_norm(self$input_dim)
    }

    self$rff <- nn_random_fourier_features(
      input_dim = self$input_dim,
      num_grids = self$num_grids,
      dropout = dropout,
      activation_expectation = activation_expectation
    )

    # Feature-wise learnable gates.
    # The base GELU branch starts dominant.
    # The Fourier branch starts small for stable early training.
    self$base_scale <- torch::nn_parameter(
      torch::torch_ones(self$input_dim)
    )

    self$fourier_scale <- torch::nn_parameter(
      torch::torch_full(c(self$input_dim), fourier_init_scale)
    )

    self$linear <- torch::nn_linear(self$input_dim, self$output_dim)
  },

  forward = function(x) {
    if (length(x$shape) != 2) {
      stop("`x` must be a 2D tensor with shape [batch_size, input_dim].",
           call. = FALSE)
    }

    x_norm <- if (self$use_layernorm) {
      self$layernorm(x)
    } else {
      x
    }

    base_part <- torch::nnf_gelu(x)
    fourier_part <- self$rff(x_norm)

    activated <- self$base_scale * base_part +
      self$fourier_scale * fourier_part

    self$linear(activated)
  }
)

torch_available <- function() {
  tryCatch(
    {
      torch::torch_manual_seed(1)
      invisible(torch::torch_tensor(0))
      TRUE
    },
    error = function(e) FALSE
  )
}

skip_if_no_torch <- function() {
  testthat::skip_if_not(
    torch_available(),
    "torch backend/Lantern is not available"
  )
}

test_that("stacked KAF network returns expected output shape", {
  skip_if_no_torch()
  torch::torch_manual_seed(123)

  model <- nn_kaf(
    layers = c(4, 16, 16, 1),
    num_grids = 8
  )

  x <- torch::torch_randn(32, 4)
  y <- model(x)

  expect_equal(as.integer(y$shape), c(32L, 1L))
})


test_that("stacked KAF network supports backpropagation", {
  skip_if_no_torch()
  torch::torch_manual_seed(123)

  model <- nn_kaf(
    layers = c(4, 16, 16, 1),
    num_grids = 8
  )

  x <- torch::torch_randn(32, 4)
  y <- model(x)

  loss <- y$mean()
  loss$backward()

  expect_false(is.null(model$kaf_layers[[1]]$base_scale$grad))
  expect_false(is.null(model$kaf_layers[[1]]$fourier_scale$grad))
  expect_false(is.null(model$kaf_layers[[1]]$linear$weight$grad))
})


test_that("stacked KAF network rejects invalid layer specifications", {
  skip_if_no_torch()
  expect_error(
    nn_kaf(layers = c(4)),
    "`layers` must be a numeric vector with at least two entries"
  )

  expect_error(
    nn_kaf(layers = c(4, 0, 1)),
    "All entries in `layers` must be positive integers"
  )
})

test_that("KAF layer returns expected output shape", {
  skip_if_no_torch()
  torch::torch_manual_seed(123)

  layer <- nn_kaf_layer(
    input_dim = 4,
    output_dim = 3,
    num_grids = 8
  )

  x <- torch::torch_randn(10, 4)
  y <- layer(x)

  expect_equal(as.integer(y$shape), c(10L, 3L))
})


test_that("KAF layer supports backpropagation through both branches", {
  skip_if_no_torch()
  torch::torch_manual_seed(123)

  layer <- nn_kaf_layer(
    input_dim = 4,
    output_dim = 3,
    num_grids = 8
  )

  x <- torch::torch_randn(10, 4)
  y <- layer(x)

  loss <- y$sum()
  loss$backward()

  expect_false(is.null(layer$base_scale$grad))
  expect_false(is.null(layer$fourier_scale$grad))
  expect_false(is.null(layer$rff$weight$grad))
  expect_false(is.null(layer$linear$weight$grad))
})


test_that("KAF layer rejects invalid dropout values", {
  skip_if_no_torch()
  expect_error(
    nn_kaf_layer(
      input_dim = 4,
      output_dim = 3,
      dropout = 1
    ),
    "`dropout` must be in \\[0, 1\\)"
  )
})

test_that("random Fourier features preserve batch size and input dimension", {
  skip_if_no_torch()
  torch::torch_manual_seed(123)

  layer <- nn_random_fourier_features(
    input_dim = 4,
    num_grids = 8
  )

  x <- torch::torch_randn(10, 4)
  y <- layer(x)

  expect_equal(as.integer(y$shape), c(10L, 4L))
})


test_that("random Fourier features support backpropagation", {
  skip_if_no_torch()
  torch::torch_manual_seed(123)

  layer <- nn_random_fourier_features(
    input_dim = 4,
    num_grids = 8
  )

  x <- torch::torch_randn(10, 4)
  y <- layer(x)

  loss <- y$sum()
  loss$backward()

  expect_false(is.null(layer$weight$grad))
  expect_false(is.null(layer$bias$grad))
  expect_false(is.null(layer$combination$weight$grad))
})

#' Random Fourier Features Layer
#'
#' Torch module implementing trainable random Fourier features for
#' Kolmogorov-Arnold Fourier Networks.
#'
#' @param input_dim Integer. Input dimension.
#' @param num_grids Integer. Number of Fourier frequencies.
#' @param dropout Numeric. Dropout probability.
#' @param activation_expectation Numeric. Scaling constant for initialization.
#'
#' @return A torch nn_module.
#'
#' @export
nn_random_fourier_features <- torch::nn_module(
  "nn_random_fourier_features",

  initialize = function(input_dim,
                        num_grids = 8,
                        dropout = 0,
                        activation_expectation = 1.64) {
    self$input_dim <- input_dim
    self$num_grids <- num_grids

    var_w <- 1 / (input_dim * activation_expectation)

    self$weight <- torch::nn_parameter(
      torch::torch_randn(input_dim, num_grids) * sqrt(var_w)
    )

    self$bias <- torch::nn_parameter(
      torch::torch_empty(num_grids)$uniform_(0, 2 * pi)
    )

    self$dropout <- torch::nn_dropout(p = dropout)
    self$combination <- torch::nn_linear(2 * num_grids, input_dim)
  },

  forward = function(x) {
    projection <- torch::torch_matmul(x, self$weight) + self$bias

    features <- torch::torch_cat(
      list(
        torch::torch_cos(projection),
        torch::torch_sin(projection)
      ),
      dim = -1
    )

    features <- self$dropout(features)
    self$combination(features)
  }
)

---
title: "Getting started with rkaf"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Getting started with rkaf}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  message = FALSE,
  warning = FALSE
)

rkaf_torch_available <- tryCatch(
  {
    torch::torch_manual_seed(123)
    invisible(torch::torch_tensor(0))
    TRUE
  },
  error = function(e) FALSE
)

knitr::opts_chunk$set(eval = rkaf_torch_available)
```

```{r torch-unavailable, echo = FALSE, eval = !rkaf_torch_available}
knitr::asis_output(
  "> Torch/Lantern is not available in this build environment, so the code chunks are shown but not executed."
)
```

# Overview

`rkaf` provides Kolmogorov-Arnold Fourier Networks for R users through the
`torch` backend.

The package supports:

- regression
- binary classification
- multiclass classification
- formula and matrix interfaces
- mini-batch training
- validation splits
- early stopping
- automatic standardization
- best-model restoration
- KAF-specific diagnostics

This vignette gives a quick tour of the main workflow.

```{r setup}
library(rkaf)

set.seed(123)
torch::torch_manual_seed(123)
```

# Regression with the matrix interface

We first fit a KAF model to a synthetic one-dimensional function with both
low-frequency and high-frequency structure.

```{r regression-data}
x <- as.matrix(seq(-1, 1, length.out = 128))

y <- sin(8 * pi * x) +
  0.35 * cos(3 * pi * x) +
  0.15 * x^2
```

```{r regression-fit}
fit <- kaf_fit(
  x = x,
  y = y,
  hidden = c(256, 256),
  num_grids = 32,
  use_layernorm = FALSE,
  epochs = 1000,
  lr = 1e-3,
  standardize_x = FALSE,
  standardize_y = TRUE,
  fourier_init_scale = 5e-2,
  restore_best = TRUE,
  verbose = FALSE,
  seed = 123
)

fit
```

```{r regression-predict}
pred <- predict(fit, x)

head(data.frame(
  observed = round(as.numeric(y), 3),
  predicted = round(pred, 3)
))
```

```{r regression-plot, fig.width=7, fig.height=4}
plot(
  x,
  y,
  type = "l",
  lwd = 2,
  xlab = "x",
  ylab = "f(x)",
  main = "KAF regression fit"
)

lines(x, pred, lwd = 2, lty = 2)

legend(
  "topright",
  legend = c("Observed", "Predicted"),
  lty = c(1, 2),
  lwd = 2,
  bty = "n"
)
```

# Regression with the formula interface

For tabular data, `rkaf` also supports a formula interface.

```{r formula-regression}
fit_mtcars <- kaf_fit_formula(
  mpg ~ wt + hp + cyl,
  data = mtcars,
  hidden = c(32, 32),
  num_grids = 16,
  epochs = 200,
  verbose = FALSE,
  seed = 123
)

fit_mtcars
```

```{r formula-predict}
mtcars_pred <- predict(fit_mtcars, mtcars)

head(data.frame(
  observed = mtcars$mpg,
  predicted = round(mtcars_pred, 2)
))
```

# Binary classification

If the response is a factor with two classes, `rkaf` automatically treats the
problem as binary classification.

```{r binary-data}
df <- mtcars

df$high_mpg <- factor(
  ifelse(df$mpg > median(df$mpg), "yes", "no"),
  levels = c("no", "yes")
)
```

```{r binary-fit}
fit_binary <- kaf_fit_formula(
  high_mpg ~ wt + hp + cyl,
  data = df,
  hidden = c(32, 32),
  num_grids = 16,
  epochs = 200,
  verbose = FALSE,
  seed = 123
)

fit_binary
```

Predicted probabilities and classes:

```{r binary-prob}
prob_binary <- predict(fit_binary, df, type = "prob")
class_binary <- predict(fit_binary, df, type = "class")

head(data.frame(
  observed = df$high_mpg,
  prob_yes = round(prob_binary, 3),
  predicted = class_binary
))
```

Confusion matrix

```{r binary-confusion-matrix}
table(
  observed = df$high_mpg,
  predicted = class_binary
)
```

Raw logits:

```{r binary-link}
head(predict(fit_binary, df, type = "link"))
```

# Multiclass classification

If the response is a factor with more than two classes, `rkaf` fits a multiclass
classifier.

```{r multiclass-fit}
fit_iris <- kaf_fit_formula(
  Species ~ .,
  data = iris,
  hidden = c(32, 32),
  num_grids = 16,
  epochs = 300,
  verbose = FALSE,
  seed = 123
)

fit_iris
```

Confusion matrix

```{r multiclass-confusion-matrix}
class_iris <- predict(fit_iris, iris, type = "class")

table(
  observed = iris$Species,
  predicted = class_iris
)
```

Class probabilities:

```{r multiclass-prob}
head(round(predict(fit_iris, iris, type = "prob"), 3))
```

# Validation and early stopping

`kaf_fit()` supports validation splits, mini-batches, and early stopping.

```{r validation-fit, eval=FALSE}
fit_val <- kaf_fit(
  x = x,
  y = y,
  hidden = c(64, 64),
  num_grids = 16,
  use_layernorm = FALSE,
  epochs = 300,
  lr = 5e-4,
  batch_size = 64,
  validation_split = 0.2,
  patience = 100,
  restore_best = TRUE,
  verbose = FALSE,
  seed = 123
)

plot(fit_val)
```

The fitted object stores both `train_loss_history` and
`validation_loss_history`, so users can inspect training and validation behavior
directly.

# KAF diagnostics

The KAF architecture contains a base/GELU branch and a Fourier branch. The
package exposes helper functions to inspect the learned branch scales and
Fourier parameters.

```{r diagnostics-scales}
scales <- extract_kaf_scales(fit)

head(scales)
```

```{r diagnostics-fourier}
fourier_params <- extract_fourier_params(fit, layer = 1)

head(fourier_params)
```

# Low-level torch interface

Advanced users can use the low-level torch modules directly.

```{r low-level}
model <- nn_kaf(
  layers = c(4, 16, 16, 1),
  num_grids = 8
)

x_tensor <- torch::torch_randn(10, 4)
y_tensor <- model(x_tensor)

y_tensor$shape
```

# Summary

The standard workflow is:

```{r workflow, eval=FALSE}
fit <- kaf_fit_formula(
  y ~ .,
  data = df,
  hidden = c(64, 64),
  num_grids = 16,
  validation_split = 0.2,
  patience = 30
)

predict(fit, newdata)
plot(fit)
extract_kaf_scales(fit)
```

For classification, use:

```{r classification-workflow, eval=FALSE}
predict(fit, newdata, type = "prob")
predict(fit, newdata, type = "class")
```

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nn-kaf-layer.R
\name{nn_kaf_layer}
\alias{nn_kaf_layer}
\title{Kolmogorov-Arnold Fourier Layer}
\usage{
nn_kaf_layer(
  input_dim,
  output_dim,
  num_grids = 8,
  dropout = 0,
  use_layernorm = TRUE,
  activation_expectation = 1.64,
  fourier_init_scale = 0.01
)
}
\arguments{
\item{input_dim}{Integer. Input dimension.}

\item{output_dim}{Integer. Output dimension.}

\item{num_grids}{Integer. Number of Fourier frequencies.}

\item{dropout}{Numeric. Dropout probability.}

\item{use_layernorm}{Logical. Whether to apply layer normalization before
the Fourier feature block.}

\item{activation_expectation}{Numeric. Scaling constant for Fourier
initialization.}

\item{fourier_init_scale}{Numeric. Initial scale of the Fourier component.}
}
\value{
A torch nn_module.
}
\description{
Torch module implementing one KAF layer using a hybrid GELU and
trainable Random Fourier Features activation.
}
\details{
The layer computes a feature-wise hybrid activation of the form:

\deqn{
z = \alpha \odot GELU(x) + \beta \odot RFF(x)
}

followed by a linear projection to the output dimension.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/kaf-fit.R
\name{predict.kaf_fit}
\alias{predict.kaf_fit}
\title{Predict from a fitted KAF model}
\usage{
\method{predict}{kaf_fit}(
  object,
  newdata,
  type = c("response", "prob", "class", "link"),
  threshold = 0.5,
  as_tensor = FALSE,
  ...
)
}
\arguments{
\item{object}{A fitted object returned by \code{kaf_fit()}.}

\item{newdata}{Matrix, data frame, vector, or 2D torch tensor.}

\item{type}{Character. Prediction type. \code{"response"} returns regression
predictions for regression, probabilities for classification. \code{"prob"}
returns probabilities for classification. \code{"class"} returns predicted
classes. \code{"link"} returns raw model logits/outputs.}

\item{threshold}{Numeric. Classification threshold used for binary class
predictions.}

\item{as_tensor}{Logical. If \code{TRUE}, return a torch tensor where supported.}

\item{...}{Unused.}
}
\value{
Predictions as a vector, matrix, factor, or torch tensor.
}
\description{
Predict from a fitted KAF model
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/methods.R
\name{plot.kaf_fit}
\alias{plot.kaf_fit}
\title{Plot a fitted KAF model}
\usage{
\method{plot}{kaf_fit}(x, type = c("loss", "fit"), newdata = NULL, y = NULL, ...)
}
\arguments{
\item{x}{A fitted object returned by \code{kaf_fit()}.}

\item{type}{Character. Either \code{"loss"} or \code{"fit"}.}

\item{newdata}{Optional predictors used when \code{type = "fit"}.}

\item{y}{Optional observed target values used when \code{type = "fit"}.}

\item{...}{Additional arguments passed to base plotting functions.}
}
\value{
Invisibly returns \code{x}.
}
\description{
Plot a fitted KAF model
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/methods.R
\name{print.kaf_fit}
\alias{print.kaf_fit}
\title{Print a fitted KAF model}
\usage{
\method{print}{kaf_fit}(x, ...)
}
\arguments{
\item{x}{A fitted object returned by \code{kaf_fit()}.}

\item{...}{Unused.}
}
\value{
Invisibly returns \code{x}.
}
\description{
Print a fitted KAF model
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/diagnostics.R
\name{extract_fourier_params}
\alias{extract_fourier_params}
\title{Extract KAF Fourier parameters}
\usage{
extract_fourier_params(object, layer = NULL)
}
\arguments{
\item{object}{A fitted object returned by \code{kaf_fit()}, or a raw \code{nn_kaf}
torch module.}

\item{layer}{Optional integer. If supplied, only extract parameters from this
layer.}
}
\value{
A data frame with Fourier weights and biases.
}
\description{
Extracts the trainable Random Fourier Feature weights and biases from a KAF
model.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nn-rff.R
\name{nn_random_fourier_features}
\alias{nn_random_fourier_features}
\title{Random Fourier Features Layer}
\usage{
nn_random_fourier_features(
  input_dim,
  num_grids = 8,
  dropout = 0,
  activation_expectation = 1.64
)
}
\arguments{
\item{input_dim}{Integer. Input dimension.}

\item{num_grids}{Integer. Number of Fourier frequencies.}

\item{dropout}{Numeric. Dropout probability.}

\item{activation_expectation}{Numeric. Scaling constant for initialization.}
}
\value{
A torch nn_module.
}
\description{
Torch module implementing trainable random Fourier features for
Kolmogorov-Arnold Fourier Networks.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/diagnostics.R
\name{extract_kaf_scales}
\alias{extract_kaf_scales}
\title{Extract KAF branch scales}
\usage{
extract_kaf_scales(object)
}
\arguments{
\item{object}{A fitted object returned by \code{kaf_fit()}, or a raw \code{nn_kaf}
torch module.}
}
\value{
A data frame with one row per layer-feature pair.
}
\description{
Extracts the learned GELU/base and Fourier branch scales from each KAF layer.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nn-kaf.R
\name{nn_kaf}
\alias{nn_kaf}
\title{Kolmogorov-Arnold Fourier Network}
\usage{
nn_kaf(
  layers,
  num_grids = 8,
  dropout = 0,
  use_layernorm = TRUE,
  activation_expectation = 1.64,
  fourier_init_scale = 0.01
)
}
\arguments{
\item{layers}{Integer vector. Network architecture, including input and
output dimensions. For example, \code{c(10, 64, 64, 1)}.}

\item{num_grids}{Integer. Number of Fourier frequencies per KAF layer.}

\item{dropout}{Numeric. Dropout probability.}

\item{use_layernorm}{Logical. Whether to apply layer normalization before
the Fourier feature block.}

\item{activation_expectation}{Numeric. Scaling constant for Fourier
initialization.}

\item{fourier_init_scale}{Numeric. Initial scale of the Fourier component.}
}
\value{
A torch nn_module.
}
\description{
Torch module implementing a stacked Kolmogorov-Arnold Fourier Network.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/formula.R
\name{kaf_fit_formula}
\alias{kaf_fit_formula}
\title{Fit a KAF model using an R formula}
\usage{
kaf_fit_formula(
  formula,
  data,
  include_intercept = FALSE,
  na.action = stats::na.omit,
  ...
)
}
\arguments{
\item{formula}{A model formula, such as \code{y ~ x1 + x2}.}

\item{data}{A data frame.}

\item{include_intercept}{Logical. Whether to keep the model-matrix intercept
column. Defaults to \code{FALSE}.}

\item{na.action}{Missing-value handling function passed to \code{model.frame()}.}

\item{...}{Additional arguments passed to \code{kaf_fit()}.}
}
\value{
An object of class \code{"kaf_fit"}.
}
\description{
Fits a Kolmogorov-Arnold Fourier Network using a formula and data frame.
This is a convenience wrapper around \code{kaf_fit()} for regression tasks.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/kaf-fit.R
\name{kaf_fit}
\alias{kaf_fit}
\title{Fit a Kolmogorov-Arnold Fourier Network}
\usage{
kaf_fit(
  x,
  y,
  task = c("auto", "regression", "binary", "multiclass"),
  hidden = c(64, 64),
  num_grids = 16,
  dropout = 0,
  use_layernorm = TRUE,
  fourier_init_scale = 0.01,
  epochs = 1000,
  lr = 0.001,
  batch_size = NULL,
  shuffle = TRUE,
  validation_split = 0,
  x_val = NULL,
  y_val = NULL,
  weight_decay = 0,
  standardize_x = TRUE,
  standardize_y = NULL,
  patience = NULL,
  verbose = TRUE,
  print_every = 100,
  seed = NULL,
  restore_best = TRUE,
  min_delta = 0
)
}
\arguments{
\item{x}{Matrix, data frame, vector, or 2D torch tensor of predictors.}

\item{y}{Vector, matrix, data frame, or torch tensor of targets.}

\item{task}{Character. One of \code{"auto"}, \code{"regression"}, \code{"binary"}, or
\code{"multiclass"}. With \code{"auto"}, factor, character, and logical targets are
treated as classification; numeric targets are treated as regression.}

\item{hidden}{Integer vector. Hidden layer sizes.}

\item{num_grids}{Integer. Number of Fourier frequencies per KAF layer.}

\item{dropout}{Numeric. Dropout probability.}

\item{use_layernorm}{Logical. Whether to apply layer normalization.}

\item{fourier_init_scale}{Numeric. Initial scale of the Fourier branch.}

\item{epochs}{Integer. Maximum number of training epochs.}

\item{lr}{Numeric. Learning rate.}

\item{batch_size}{Optional integer. Mini-batch size. If \code{NULL}, full-batch
training is used.}

\item{shuffle}{Logical. Whether to shuffle training rows each epoch.}

\item{validation_split}{Numeric in \verb{[0, 1)}. Fraction of rows to reserve for
validation. Ignored if \code{x_val} and \code{y_val} are supplied.}

\item{x_val}{Optional validation predictors.}

\item{y_val}{Optional validation targets.}

\item{weight_decay}{Numeric. Adam weight decay.}

\item{standardize_x}{Logical. Whether to standardize predictors using the
training-set mean and standard deviation.}

\item{standardize_y}{Logical or \code{NULL}. Whether to standardize regression
targets using the training-set mean and standard deviation. If \code{NULL},
targets are standardized for regression and not standardized for
classification. Predictions are automatically transformed back to the
original target scale.}

\item{patience}{Optional integer. Number of epochs without improvement before
early stopping.}

\item{verbose}{Logical. Whether to print progress.}

\item{print_every}{Integer. Print frequency.}

\item{seed}{Optional integer random seed.}

\item{restore_best}{Logical. Whether to restore the best observed model state
after training.}

\item{min_delta}{Numeric. Minimum loss improvement required to update the
best model state.}
}
\value{
An object of class \code{"kaf_fit"}.
}
\description{
Fits a KAF model for regression using mean squared error loss.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/diagnostics.R
\name{plot_kaf_scales}
\alias{plot_kaf_scales}
\title{Plot KAF branch scales}
\usage{
plot_kaf_scales(object, layer = 1, type = c("ratio", "branch"), ...)
}
\arguments{
\item{object}{A fitted object returned by \code{kaf_fit()}, or a raw \code{nn_kaf}
torch module.}

\item{layer}{Integer. Layer to inspect.}

\item{type}{Character. Either \code{"ratio"} or \code{"branch"}.}

\item{...}{Additional arguments passed to base plotting functions.}
}
\value{
Invisibly returns the extracted scale data.
}
\description{
Visualizes the learned base/GELU and Fourier branch scales for a selected KAF
layer.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rkaf-package.R
\docType{package}
\name{rkaf-package}
\alias{rkaf}
\alias{rkaf-package}
\title{rkaf: Kolmogorov-Arnold Fourier Networks in R}
\description{
\code{rkaf} provides an R implementation of Kolmogorov-Arnold Fourier Networks
using the \code{torch} backend.
}
\details{
The package supports:
\itemize{
\item regression
\item binary classification
\item multiclass classification
\item formula and matrix interfaces
\item mini-batch training
\item validation splits
\item early stopping
\item automatic standardization
\item best-model restoration
\item KAF-specific diagnostics
}
}
\seealso{
Useful links:
\itemize{
  \item \url{https://github.com/gsidoine/rkaf}
  \item \url{https://gsidoine.github.io/rkaf/}
  \item Report bugs at \url{https://github.com/gsidoine/rkaf/issues}
}

}
\author{
\strong{Maintainer}: Guillaume Sidoine \email{gjh.sidoine@gmail.com}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/kaf.R
\name{kaf}
\alias{kaf}
\title{Create a Kolmogorov-Arnold Fourier Network}
\usage{
kaf(
  input_dim,
  output_dim = 1,
  hidden = c(64, 64),
  num_grids = 16,
  dropout = 0,
  use_layernorm = TRUE,
  activation_expectation = 1.64,
  fourier_init_scale = 0.01
)
}
\arguments{
\item{input_dim}{Integer. Number of input features.}

\item{output_dim}{Integer. Number of output dimensions.}

\item{hidden}{Integer vector. Hidden layer sizes.}

\item{num_grids}{Integer. Number of Fourier frequencies per KAF layer.}

\item{dropout}{Numeric. Dropout probability.}

\item{use_layernorm}{Logical. Whether to apply layer normalization.}

\item{activation_expectation}{Numeric. Scaling constant for Fourier initialization.}

\item{fourier_init_scale}{Numeric. Initial scale of the Fourier branch.}
}
\value{
A torch KAF network.
}
\description{
Convenience wrapper around \code{nn_kaf()} using a more user-facing API.
}
